from __future__ import with_statement
from django.contrib.auth.models import AnonymousUser
from django.contrib.sessions.middleware import SessionMiddleware
from django.core.urlresolvers import reverse
from django.http import QueryDict
from django.test.client import Client, RequestFactory
from django_facebook import exceptions as facebook_exceptions, \
    settings as facebook_settings, signals
from django_facebook.api import get_facebook_graph, FacebookUserConverter, \
    get_persistent_graph
from django_facebook.auth_backends import FacebookBackend
from django_facebook.connect import _register_user, connect_user, \
    CONNECT_ACTIONS
from django_facebook.middleware import FacebookCanvasMiddleWare
from django_facebook.test_utils.mocks import RequestMock
from django_facebook.test_utils.testcases import FacebookTest, LiveFacebookTest
from django_facebook.utils import cleanup_oauth_url, get_profile_model, \
    ScriptRedirect, get_user_model, get_user_attribute, try_get_profile, \
    get_instance_for_attribute, update_user_attributes, get_registration_backend
from functools import partial
from mock import Mock, patch
from open_facebook.api import FacebookConnection, FacebookAuthorization, \
    OpenFacebook
from open_facebook.exceptions import FacebookSSLError, FacebookURLError
import logging
import mock
from django.utils import six
from django_facebook.models import OpenGraphShare
from django.contrib.contenttypes.models import ContentType
from open_facebook.exceptions import FacebookUnreachable, OAuthException

logger = logging.getLogger(__name__)
__doctests__ = ['django_facebook.api']


class BaseDecoratorTest(FacebookTest):

    def setUp(self):
        FacebookTest.setUp(self)
        from django_facebook.decorators import facebook_required
        self.decorator = facebook_required
        self.decorator_name = 'FacebookRequired'

    def test_naming(self):
        self.assertEqual(self.decorator.__name__, self.decorator_name)

    def test_wrapping(self):
        '''
        Verify that the decorator wraps the original function
        '''

        @self.decorator
        def myfunc(request):
            '''docs'''
            pass
        self.assertEqual(myfunc.__doc__, 'docs')
        self.assertEqual(myfunc.__name__, 'myfunc')

        @self.decorator()
        def myfunc2(request):
            '''docs2'''
            pass
        self.assertEqual(myfunc2.__doc__, 'docs2')
        self.assertEqual(myfunc2.__name__, 'myfunc2')


class DecoratorTest(BaseDecoratorTest):

    '''
    Verify that the lazy and facebook_required decorator work as expected

    Facebook required decorator
        If you have the permissions proceed
        Else show the login screen
            If you allow, proceed
            If you click cancel ...

    Facebook required lazy
        Proceed
        Upon OAuthException, go to login screen
            If you allow proceed
            If you click cancel ...
    '''

    def setUp(self):
        BaseDecoratorTest.setUp(self)
        self.url = reverse('facebook_decorator_example')
        target_url = r'''https://www.facebook.com/dialog/oauth?scope=email%2Cuser_about_me%2Cuser_birt
            hday%2Cuser_website&redirect_uri=http%3A%2F%2Ftestserver%2Ffacebook%2Fdecorator_
            example%2F%3Fattempt%3D1&client_id=215464901804004
        '''.replace(' ', '').replace('\n', '')
        self.target_url = target_url
        from django_facebook.decorators import facebook_required
        self.decorator = facebook_required

    def test_decorator_not_authenticated(self):
        '''
        We should redirect to Facebook oauth dialog
        '''
        response = self.client.get(self.url, follow=True)
        if six.PY3:
            self.assertEqual(response.redirect_chain[0][1], 302)
        else:
            self.assertRedirects(
                response, self.target_url, target_status_code=404)

    def test_decorator_authenticated(self):
        '''
        Here we fake that we have permissions
        This should enter the view and in this test return "authorized"
        '''
        self.mock_authenticated()
        response = self.client.get(self.url, follow=True)
        if type(response.content) is six.binary_type:
            self.assertEqual(response.content.decode(), 'authorized')
        else:
            self.assertEqual(response.content, 'authorized')

    def test_decorator_denied(self):
        '''
        Here the users denies our app. Facebook adds this in the url
        attempt=1&error_reason=user_denied&error=access_denied&error_description=The+user+denied+your+request.
        '''
        query_dict_string = 'attempt=1&error_reason=user_denied&error=access_denied&error_description=The+user+denied+your+request.'
        get = QueryDict(query_dict_string, True)
        denied_url = '%s?%s' % (self.url, get.urlencode())
        response = self.client.get(denied_url, follow=True)
        if type(response.content) is six.binary_type:
            self.assertEqual(response.content.decode(), 'user denied or error')
        else:
            self.assertEqual(response.content, 'user denied or error')


class ScopedDecoratorTest(DecoratorTest):

    '''
    Tests the more complicated but faster lazy decorator
    '''

    def setUp(self):
        DecoratorTest.setUp(self)
        self.url = reverse('facebook_decorator_example_scope')
        target_url = r'https://www.facebook.com/dialog/oauth?scope=publish_actions%2Cuser_status&redirect_uri=http%3A%2F%2Ftestserver%2Ffacebook%2Fdecorator_example_scope%2F%3Fattempt%3D1&client_id=215464901804004'
        self.target_url = target_url

    def test_type_error(self):
        self.mock_authenticated()

        @self.decorator
        def myview(request, graph):
            def inner(a, b):
                pass
            inner(1, 2, c='nono')

        to_fail = partial(myview, self.request)
        try:
            to_fail()
        except TypeError as e:
            right_error = "inner() got an unexpected keyword argument 'c'"
            self.assertEqual(str(e), right_error)


class LazyDecoratorTest(DecoratorTest):

    '''
    Tests the more complicated but faster lazy decorator
    '''

    def setUp(self):
        DecoratorTest.setUp(self)
        self.url = reverse('facebook_lazy_decorator_example')
        target_url = r'''https://www.facebook.com/dialog/oauth?scope=email%2Cuser_about_me%2Cuser_birt
            hday%2Cuser_website&redirect_uri=http%3A%2F%2Ftestserver%2Ffacebook%2Flazy_decorator_
            example%2F%3Fattempt%3D1&client_id=215464901804004
        '''.replace(' ', '').replace('\n', '')
        self.target_url = target_url
        from django_facebook.decorators import facebook_required_lazy
        self.decorator = facebook_required_lazy
        self.decorator_name = 'FacebookRequiredLazy'


class GraphAccessTest(FacebookTest):

    def test_get_persistent(self):
        graph = get_persistent_graph(self.request)
        # fake that we are authenticated and have a facebook graph
        with patch.object(self.request, 'facebook'):
            self.request.user = get_user_model().objects.all()[:1][0]
            graph = get_persistent_graph(self.request)


class ConnectViewTest(FacebookTest):

    def setUp(self):
        FacebookTest.setUp(self)

        # According to django/test/testcase.py, the scheme and
        # domain are not included on redirect_chain on django 1.9
        self.default_url = \
            facebook_settings.FACEBOOK_LOGIN_DEFAULT_REDIRECT
        self.url = reverse('facebook_connect')
        self.example_url = reverse('facebook_example')

    def test_connect_redirect(self):
        '''
        The redirect flow for facebook works as follows

        - request the decorated url, /facebook/connect/
        - the decorator (facebook_required) redirect the user to the oauth url
        - after accepting the auth dialog facebook redirects us to the next url
        '''
        # STEP 1, verify that we redirect to facebook with the correct details
        response = self.client.post(
            self.url, next=self.example_url, follow=True)
        redirect_url = response.redirect_chain[0][0]
        oauth_url = 'https://www.facebook.com/dialog/oauth?scope=email%2Cuser_about_me%2Cuser_birthday%2Cuser_website&redirect_uri=http%3A%2F%2Ftestserver%2Ffacebook%2Fconnect%2F%3Fattempt%3D1&client_id=215464901804004'
        if six.PY3:
            self.assertEqual(response.redirect_chain[0][1], 302)
        else:
            self.assertEqual(redirect_url, oauth_url)

    def test_connect_redirect_authenticated(self):
        # Meanwhile at Facebook they redirect the request
        # STEP 2 Authenticated, verify that the connect view redirects to the
        # example
        self.mock_authenticated()
        accepted_url = self.url + \
            '?attempt=1&client_id=215464901804004&next=bla&register_next=%s' % self.example_url
        response = self.client.get(accepted_url, follow=True)
        redirect_url = response.redirect_chain[0][0]
        self.assertIn(self.example_url, redirect_url)

        # Verify that login_next works
        accepted_url = self.url + \
            '?attempt=1&client_id=215464901804004&next=bla&login_next=%s' % self.example_url
        response = self.client.get(accepted_url, follow=True)
        redirect_url = response.redirect_chain[0][0]
        self.assertIn(self.example_url, redirect_url)

    def test_connect_redirect_default(self):
        # Now try without next
        self.mock_authenticated()
        accepted_url = self.url + '?attempt=1&client_id=215464901804004'
        response = self.client.get(accepted_url, follow=True)
        redirect_url = response.redirect_chain[0][0]
        self.assertIn(self.default_url, redirect_url)

    def test_connect_redirect_not_authenticated(self):
        # Meanwhile at Facebook they redirect the request
        # STEP 2 Not Authenticated, verify that the connect view redirects to
        # the example
        accepted_url = self.url + \
            '?attempt=1&client_id=215464901804004&next=%s' % self.example_url
        response = self.client.get(accepted_url, follow=True)
        redirect_url = response.redirect_chain[0][0]
        error_url = self.example_url + '?fb_error_or_cancel=1'
        self.assertIn(error_url, redirect_url)

        # Verify that error next also works
        accepted_url = self.url + \
            '?attempt=1&client_id=215464901804004&next=bla&error_next=%s' % self.example_url
        response = self.client.get(accepted_url, follow=True)
        redirect_url = response.redirect_chain[0][0]
        error_url = self.example_url + '?fb_error_or_cancel=1'
        self.assertIn(error_url, redirect_url)

    def test_connect(self):
        '''
        Test if we can do logins
        django_facebook.connect.connect_user
        '''
        user = get_user_model().objects.all()[:1][0]
        url = self.url
        example_url = reverse('facebook_example')

        # test registration flow
        with patch('django_facebook.views.connect_user', return_value=(CONNECT_ACTIONS.REGISTER, user)) as wrapped_connect:
            post_data = dict(
                access_token='short_username',
                next='%s?register=1' % example_url,
            )
            response = self.client.post(url, post_data, follow=True)
            self.assertEqual(wrapped_connect.call_count, 1)
            self.assertIn('register', response.redirect_chain[0][0])
            self.assertEqual(response.status_code, 200)

        # user register next instead of next
        with patch('django_facebook.views.connect_user', return_value=(CONNECT_ACTIONS.REGISTER, user)) as wrapped_connect:
            post_data = dict(
                access_token='short_username',
                register_next='%s?register=1' % example_url
            )
            response = self.client.post(url, post_data, follow=True)
            self.assertEqual(wrapped_connect.call_count, 1)
            self.assertIn('register', response.redirect_chain[0][0])
            self.assertEqual(response.status_code, 200)

        # test login
        with patch('django_facebook.views.connect_user', return_value=(CONNECT_ACTIONS.LOGIN, user)) as wrapped_connect:
            post_data = dict(
                access_token='short_username',
                next='%s?loggggg=1' % example_url,
            )
            response = self.client.post(url, post_data, follow=True)
            self.assertEqual(wrapped_connect.call_count, 1)
            self.assertIn('?loggggg=1', response.redirect_chain[0][0])
            self.assertEqual(response.status_code, 200)

        # test connect
        with patch('django_facebook.views.connect_user', return_value=(CONNECT_ACTIONS.CONNECT, user)) as wrapped_connect:
            post_data = dict(
                access_token='short_username',
                next='%s?loggggg=1' % example_url
            )
            response = self.client.post(url, post_data, follow=True)
            self.assertEqual(wrapped_connect.call_count, 1)
            assert '?loggggg=1' in response.redirect_chain[0][0]
            self.assertEqual(response.status_code, 200)

        # test connect
        from django_facebook import exceptions as facebook_exceptions
        profile_error = facebook_exceptions.IncompleteProfileError()
        profile_error.form = None
        with patch('django_facebook.views.connect_user', return_value=(CONNECT_ACTIONS.REGISTER, user), side_effect=profile_error) as wrapped_connect:
            post_data = dict(access_token='short_username',
                             next='%s?loggggg=1' % example_url)
            response = self.client.post(url, post_data, follow=True)
            self.assertEqual(wrapped_connect.call_count, 1)
            self.assertEqual(response.status_code, 200)
            self.assertTrue(response.context)
            template = self.get_response_template(response)
            backend = get_registration_backend()
            assert template.name in backend.get_registration_template()

    def test_slow_connect(self):
        '''
        Test if we can do logins
        django_facebook.connect.connect_user
        '''
        url = reverse('facebook_connect')
        example_url = reverse('facebook_example')

        # test super slow Facebook
        errors = [FacebookSSLError(), FacebookURLError(
            '<urlopen error _ssl.c:489: The handshake operation timed out>')]
        for error in errors:
            with patch('django_facebook.views.get_instance_for') as converter:
                instance = converter.return_value
                instance.is_authenticated = Mock(side_effect=error)
                post_data = dict(
                    access_token='short_username',
                    next='%s?loggggg=1' % example_url
                )
                response = self.client.post(url, post_data, follow=True)
                self.assertEqual(instance.is_authenticated.call_count, 1)
                self.assertTrue(response.context)
                assert '?loggggg=1' in response.redirect_chain[0][0]

    def get_response_template(self, response):
        if hasattr(response, 'template'):
            templates = [response.template]
        else:
            templates = response.templates
        template = templates[0]
        return template


class TestUserTest(LiveFacebookTest):

    def test_create_test_user(self):
        # Also, somehow unittest.skip doesnt work with travis ci?
        return 'Skipping since you might have created test users manually, lets not delete them :)'
        # start by clearing out our test users (maybe this isnt safe to use in testing)
        # if other people create test users manualy this could be annoying
        app_access_token = FacebookAuthorization.get_cached_app_access_token()
        FacebookAuthorization.delete_test_users(app_access_token)
        # the permissions for which we want a test user
        permissions = ['email', 'publish_actions']
        # gets the test user object
        test_user = FacebookAuthorization.get_or_create_test_user(
            app_access_token, permissions)
        graph = test_user.graph()
        me = graph.me()
        assert me


class ExtendTokenTest(LiveFacebookTest):

    def test_extend_token(self):
        return 'this doesnt work in travis, but locally its fine... weird'
        app_access_token = FacebookAuthorization.get_cached_app_access_token()
        test_user = FacebookAuthorization.get_or_create_test_user(
            app_access_token)
        access_token = test_user.access_token
        results = FacebookAuthorization.extend_access_token(access_token)
        if 'access_token' not in results:
            raise ValueError('we didnt get a fresh token')


class OpenGraphShareTest(FacebookTest):

    def setUp(self):
        FacebookTest.setUp(self)
        user_url = 'http://www.fashiolista.com/style/neni/'
        kwargs = dict(item=user_url)
        user = get_user_model().objects.all()[:1][0]
        profile = try_get_profile(user)
        user_or_profile = get_instance_for_attribute(
            user, profile, 'facebook_open_graph')
        user_or_profile.facebook_open_graph = True
        user_or_profile.save()

        some_content_type = ContentType.objects.all()[:1][0]
        share = OpenGraphShare.objects.create(
            user_id=user.id,
            facebook_user_id=13123123,
            action_domain='fashiolista:follow',
            content_type=some_content_type,
            object_id=user.id,
        )
        share.set_share_dict(kwargs)
        share.save()
        self.share = share
        self.share_details = user, profile, share

    def test_follow_og_share(self):
        user_url = 'http://www.fashiolista.com/style/neni/'
        kwargs = dict(item=user_url)
        user = get_user_model().objects.all()[:1][0]
        from django.contrib.contenttypes.models import ContentType
        some_content_type = ContentType.objects.all()[:1][0]
        share = OpenGraphShare.objects.create(
            user_id=user.id,
            facebook_user_id=13123123,
            action_domain='fashiolista:follow',
            content_type=some_content_type,
            object_id=user.id,
        )
        share.set_share_dict(kwargs)
        share.save()
        share.send()

    def test_follow_og_share_error(self):
        '''
        A normal OpenFacebook exception, shouldnt reset the new token required

        However an OAuthException should set new_token_required to True,
        But only if we are indeed failing has_permissions(['publish_actions'])
        '''
        # utility function for testing purposes
        def test_send(error, expected_error_message, expected_new_token, has_permissions=False):
            user, profile, share = self.share_details
            update_user_attributes(
                user, profile, dict(new_token_required=False), save=True)
            with mock.patch('open_facebook.api.OpenFacebook') as mocked:
                instance = mocked.return_value
                instance.set = Mock(side_effect=error)
                instance.has_permissions = Mock(return_value=has_permissions)
                instance.access_token = get_user_attribute(
                    user, profile, 'access_token')
                share.send(graph=instance)
                self.assertEqual(share.error_message, expected_error_message)
                self.assertFalse(share.completed_at)
                user = get_user_model().objects.get(id=user.id)
                if profile:
                    profile = get_profile_model().objects.get(id=profile.id)
                new_token_required = get_user_attribute(
                    user, profile, 'new_token_required')
                self.assertEqual(new_token_required, expected_new_token)

        # test with a basic exception, this should reset the new_token_required
        test_send(
            error=FacebookUnreachable('broken'),
            expected_error_message='FacebookUnreachable(\'broken\',)',
            expected_new_token=False,
            has_permissions=False,
        )
        # now try with an oAuthException and no permissions
        # this should set new_token_required to true
        test_send(
            error=OAuthException('permissions'),
            expected_error_message="OAuthException('permissions',)",
            expected_new_token=True,
            has_permissions=False,
        )
        # now an oAuthException, but we already have the permissions
        # this means we shouldnt set new_token_required to True
        test_send(
            error=OAuthException('permissions'),
            expected_error_message="OAuthException('permissions',)",
            expected_new_token=False,
            has_permissions=True,
        )


class UserConnectTest(FacebookTest):

    '''
    Tests the connect user functionality
    '''

    def test_persistent_graph(self):
        request = RequestMock().get('/')
        request.session = {}
        request.user = AnonymousUser()

        graph = get_facebook_graph(request=self.request, access_token='short_username')
        FacebookUserConverter(graph)
        action, user = connect_user(self.request, facebook_graph=graph)
        self.assertEqual(action, CONNECT_ACTIONS.REGISTER)

    def test_gender_matching(self):
        request = RequestMock().get('/')
        request.session = {}
        request.user = AnonymousUser()
        graph = get_persistent_graph(request, access_token='paul')
        converter = FacebookUserConverter(graph)
        base_data = converter.facebook_profile_data()
        self.assertEqual(base_data['gender'], 'male')
        data = converter.facebook_registration_data()
        self.assertEqual(data['gender'], 'm')
        action, user = connect_user(self.request, facebook_graph=graph)
        profile = try_get_profile(user)
        gender = get_user_attribute(user, profile, 'gender')
        self.assertEqual(gender, 'm')

    def test_update_access_token(self):
        request = RequestMock().get('/')
        request.session = {}
        request.user = AnonymousUser()
        graph = get_persistent_graph(request, access_token='paul')
        action, user = connect_user(self.request, facebook_graph=graph)
        first_user_id = user.id

        # new token required should start out as False
        profile = try_get_profile(user)
        new_token_required = get_user_attribute(
            user, profile, 'new_token_required')
        self.assertEqual(new_token_required, False)

        # we manually set it to true
        update_user_attributes(
            user, profile, dict(new_token_required=True), save=True)
        if profile:
            profile = get_profile_model().objects.get(id=profile.id)
        user = get_user_model().objects.get(id=user.id)
        new_token_required = get_user_attribute(
            user, profile, 'new_token_required')
        self.assertEqual(new_token_required, True)

        # another update should however set it back to False
        request.facebook = None
        graph = get_facebook_graph(request, access_token='paul2')
        logger.info('and the token is %s', graph.access_token)
        action, user = connect_user(self.request, facebook_graph=graph)
        user = get_user_model().objects.get(id=user.id)
        self.assertEqual(user.id, first_user_id)
        if profile:
            profile = get_profile_model().objects.get(id=profile.id)
        user = get_user_model().objects.get(id=user.id)
        new_token_required = get_user_attribute(
            user, profile, 'new_token_required')
        self.assertEqual(new_token_required, False)

    def test_long_username(self):
        request = RequestMock().get('/')
        request.session = {}
        request.user = AnonymousUser()
        graph = get_persistent_graph(request, access_token='long_username')
        converter = FacebookUserConverter(graph)
        base_data = converter.facebook_registration_data()
        action, user = connect_user(self.request, facebook_graph=graph)
        self.assertEqual(len(base_data['username']), 30)
        self.assertEqual(len(user.username), 30)
        self.assertEqual(len(user.first_name), 30)
        self.assertEqual(len(user.last_name), 30)

    def test_full_connect(self):
        # going for a register, connect and login
        graph = get_facebook_graph(request=self.request, access_token='short_username')
        FacebookUserConverter(graph)
        action, user = connect_user(self.request, facebook_graph=graph)
        self.assertEqual(action, CONNECT_ACTIONS.REGISTER)
        # and now we do a login, not a connect
        action, user = connect_user(self.request, facebook_graph=graph)
        self.assertEqual(action, CONNECT_ACTIONS.LOGIN)
        self.request.GET._mutable = True
        self.request.GET['connect_facebook'] = 1
        action, user = connect_user(
            self.request, facebook_graph=graph, connect_facebook=True)
        self.assertEqual(action, CONNECT_ACTIONS.CONNECT)
        self.request.user = AnonymousUser()
        action, user = connect_user(
            self.request, facebook_graph=graph, connect_facebook=True)
        self.assertEqual(action, CONNECT_ACTIONS.LOGIN)

    def test_parallel_register(self):
        '''
        Adding some testing for the case when one person tries to register
        multiple times in the same second
        '''
        graph = get_facebook_graph(request=self.request, access_token='short_username')
        FacebookUserConverter(graph)
        action, user = connect_user(self.request, facebook_graph=graph)
        self.assertEqual(action, CONNECT_ACTIONS.REGISTER)

        self.request.user.is_authenticated = lambda: False
        with patch('django_facebook.connect.authenticate') as patched:
            return_sequence = [user, None]

            def side(*args, **kwargs):
                value = return_sequence.pop()
                return value

            patched.side_effect = side
            with patch('django_facebook.connect._register_user') as patched_register:
                patched_register.side_effect = facebook_exceptions.AlreadyRegistered(
                    'testing parallel registers')
                action, user = connect_user(self.request, facebook_graph=graph)
                self.assertEqual(action, CONNECT_ACTIONS.LOGIN)

    def test_utf8(self):
        graph = get_facebook_graph(request=self.request, access_token='unicode_string')
        facebook = FacebookUserConverter(graph)
        action, user = connect_user(self.request, facebook_graph=graph)

    def test_invalid_token(self):
        self.assertRaises(AssertionError,
                          connect_user, self.request, access_token='invalid')

    def test_no_email_registration(self):
        from django_facebook import exceptions as facebook_exceptions
        self.assertRaises(facebook_exceptions.IncompleteProfileError,
                          connect_user, self.request, access_token='no_email')

    def test_current_user(self):
        facebook = get_facebook_graph(request=self.request, access_token='tschellenbach')
        action, user = connect_user(self.request, facebook_graph=facebook)
        self.assertEqual(action, CONNECT_ACTIONS.LOGIN)

    def test_fb_update_required(self):
        def pre_update(sender, user, profile, facebook_data, **kwargs):
            user.pre_update_signal = True

        Profile = get_profile_model()
        user_model = get_user_model()
        signals.facebook_pre_update.connect(pre_update, sender=user_model)
        facebook = get_facebook_graph(request=self.request, access_token='tschellenbach')

        facebook_settings.FACEBOOK_FORCE_PROFILE_UPDATE_ON_LOGIN = True
        action, user = connect_user(self.request, facebook_graph=facebook)
        self.assertEqual(action, CONNECT_ACTIONS.LOGIN)
        self.assertTrue(hasattr(user, 'pre_update_signal'))

        facebook_settings.FACEBOOK_FORCE_PROFILE_UPDATE_ON_LOGIN = False
        action, user = connect_user(self.request, facebook_graph=facebook)
        self.assertEqual(action, CONNECT_ACTIONS.LOGIN)
        self.assertFalse(hasattr(user, 'pre_update_signal'))

    def test_new_user(self):
        facebook = get_facebook_graph(request=self.request, access_token='new_user')
        action, user = connect_user(self.request, facebook_graph=facebook)

    def test_short_username(self):
        facebook = get_facebook_graph(request=self.request, access_token='short_username')
        action, user = connect_user(self.request, facebook_graph=facebook)
        self.assertTrue(len(user.username) > 4)
        self.assertEqual(action, CONNECT_ACTIONS.REGISTER)

    def test_gender(self):
        graph = get_facebook_graph(request=self.request, access_token='new_user')
        facebook = FacebookUserConverter(graph)
        data = facebook.facebook_registration_data()
        self.assertEqual(data['gender'], 'm')

    def test_double_username(self):
        '''
        This used to give an error with duplicate usernames
        with different capitalization
        '''
        facebook = get_facebook_graph(request=self.request, access_token='short_username')
        action, user = connect_user(self.request, facebook_graph=facebook)
        user.username = 'Thierry_schellenbach'
        user.save()
        self.request.user = AnonymousUser()
        facebook = get_facebook_graph(request=self.request, access_token='same_username')
        action, new_user = connect_user(self.request, facebook_graph=facebook)
        self.assertNotEqual(user.username, new_user.username)
        self.assertNotEqual(user.id, new_user.id)

    def test_registration_form(self):
        '''
        Django_facebook should use user supplied registration form if given
        '''
        from django.conf import settings
        if settings.MODE == 'userena':
            return

        test_form = 'django_facebook.test_utils.forms.SignupForm'
        old_setting = facebook_settings.FACEBOOK_REGISTRATION_FORM
        facebook_settings.FACEBOOK_REGISTRATION_FORM = test_form
        try:
            facebook = get_facebook_graph(request=self.request, access_token='short_username')
            action, user = connect_user(self.request, facebook_graph=facebook)
            # The test form always sets username to test form
            self.assertEqual(user.username, 'Test form')
        finally:
            facebook_settings.FACEBOOK_REGISTRATION_FORM = old_setting


class SimpleRegisterViewTest(FacebookTest):

    '''
    Even the most simple views will break eventually if they are not tested
    '''

    def test_registration(self):
        pw = 'tester1234'
        data = dict(username='testertester', email='tester@testertester.com',
                    password1=pw, password2=pw)
        data['register_next'] = '/?a=bbbbcbbbb'
        response = self.client.post('/accounts/register/', data, follow=True)

        assert response.redirect_chain, 'we are expecting a redirect!'
        for url, status in response.redirect_chain:
            if 'bbbbcbbbb' in url:
                break
        else:
            raise ValueError('bbbbcbbbb isnt in %s' % response.redirect_chain)


class AuthBackend(FacebookTest):

    def test_auth_backend(self):
        # the auth backend
        backend = FacebookBackend()
        facebook = get_facebook_graph(request=self.request, access_token='new_user')
        action, user = connect_user(self.request, facebook_graph=facebook)
        facebook_email = user.email
        profile = try_get_profile(user)
        user_or_profile = get_instance_for_attribute(
            user, profile, 'facebook_id')
        facebook_id = user_or_profile.facebook_id
        auth_user = backend.authenticate(facebook_email=facebook_email)
        logger.info('%s %s %s', auth_user.email, user.email, facebook_email)
        self.assertEqual(auth_user, user)

        auth_user = backend.authenticate(facebook_id=facebook_id)
        self.assertEqual(auth_user, user)

        auth_user = backend.authenticate(facebook_id=facebook_id,
                                         facebook_email=facebook_email)
        self.assertEqual(auth_user, user)

        auth_user = backend.authenticate()
        self.assertIsNone(auth_user)


class ErrorMappingTest(FacebookTest):

    def test_mapping(self):
        from open_facebook import exceptions as open_facebook_exceptions
        raise_something = partial(FacebookConnection.raise_error, 0,
                                  "(#200) The user hasn't authorized the "
                                  "application to perform this action")
        self.assertRaises(open_facebook_exceptions.PermissionException,
                          raise_something)


class OAuthUrlTest(FacebookTest):

    def _test_equal(self, url, output):
        converted = cleanup_oauth_url(url)
        self.assertEqual(converted, output)

    def test_url(self):
        url = 'http://www.google.com/'
        output = 'http://www.google.com/'
        self._test_equal(url, output)

        url = 'http://www.google.com/?code=a'
        output = 'http://www.google.com/'
        self._test_equal(url, output)

        url = 'http://www.google.com/?code=a&b=c&d=c'
        output = 'http://www.google.com/?b=c&d=c'
        self._test_equal(url, output)


class SignalTest(FacebookTest):

    '''
    Tests that signals fire properly
    '''

    def test_user_registered_signal(self):
        # Ensure user registered, pre update and post update signals fire

        def user_registered(sender, user, facebook_data, **kwargs):
            user.registered_signal = True

        def pre_update(sender, user, profile, facebook_data, **kwargs):
            user.pre_update_signal = True

        def post_update(sender, user, profile, facebook_data, **kwargs):
            user.post_update_signal = True

        Profile = get_profile_model()
        user_model = get_user_model()
        signals.facebook_user_registered.connect(
            user_registered, sender=user_model)
        signals.facebook_pre_update.connect(pre_update, sender=user_model)
        signals.facebook_post_update.connect(post_update, sender=user_model)

        graph = get_facebook_graph(request=self.request, access_token='short_username')
        facebook = FacebookUserConverter(graph)
        user = _register_user(self.request, facebook)
        self.assertEqual(hasattr(user, 'registered_signal'), True)
        self.assertEqual(hasattr(user, 'pre_update_signal'), True)
        self.assertEqual(hasattr(user, 'post_update_signal'), True)


def fake_connect(request, access_tokon, graph):
    return ('action', 'user')


class FacebookCanvasMiddlewareTest(FacebookTest):

    def setUp(self):
        super(FacebookCanvasMiddlewareTest, self).setUp()
        self.factory = RequestFactory()
        self.middleware = FacebookCanvasMiddleWare()
        self.session_middleware = SessionMiddleware()

    def get_canvas_url(self, data={}):
        request = self.factory.post('/', data)
        request.META['HTTP_REFERER'] = 'https://apps.facebook.com/canvas/'
        self.session_middleware.process_request(request)
        return request

    def test_referer(self):
        # test empty referer
        request = self.factory.get('/')
        self.assertIsNone(self.middleware.process_request(request))
        # test referer not facebook
        request = self.factory.get('/')
        request.META['HTTP_REFERER'] = 'https://localhost:8000/'
        self.assertIsNone(self.middleware.process_request(request))
        request = self.get_canvas_url()
        response = self.middleware.process_request(request)
        self.assertIsInstance(response, ScriptRedirect)

    def test_user_denied(self):
        request = self.factory.get(
            '/?error_reason=user_denied&error=access_denied&error_description=The+user+denied+your+request.')
        request.META['HTTP_REFERER'] = 'https://apps.facebook.com/canvas/'
        response = self.middleware.process_request(request)
        self.assertIsInstance(response, ScriptRedirect)

    @patch.object(FacebookAuthorization, 'parse_signed_data')
    def test_non_auth_user(self, mocked_method=FacebookAuthorization.parse_signed_data):
        mocked_method.return_value = {}
        data = {'signed_request':
                'dXairHLF8dfUKaL7ZFXaKmTsAglg0EkyHesTLnPcPAE.eyJhbGdvcml0aG0iOiJITUFDLVNIQTI1NiIsImlzc3VlZF9hdCI6MTM1ODA2MTU1MSwidXNlciI6eyJjb3VudHJ5IjoiYnIiLCJsb2NhbGUiOiJlbl9VUyIsImFnZSI6eyJtaW4iOjIxfX19'}
        request = self.get_canvas_url(data=data)
        response = self.middleware.process_request(request)
        self.assertTrue(mocked_method.called)
        self.assertIsInstance(response, ScriptRedirect)

    @patch('django_facebook.middleware.connect_user', fake_connect)
    @patch.object(OpenFacebook, 'permissions')
    @patch.object(FacebookAuthorization, 'parse_signed_data')
    def test_auth_user(
        self, mocked_method_1=FacebookAuthorization.parse_signed_data,
            mocked_method_2=OpenFacebook.permissions):
        data = {'signed_request':
                'd7JQQIfxHgEzLIqJMeU9J5IlLg7shzPJ8DFRF55L52w.eyJhbGdvcml0aG0iOiJITUFDLVNIQTI1NiIsImV4cGlyZXMiOjEzNTgwNzQ4MDAsImlzc3VlZF9hdCI6MTM1ODA2ODU1MCwib2F1dGhfdG9rZW4iOiJBQUFGdk02MWpkT0FCQVBhWkNzR1pDM0dEVFZtdDJCWkFQVlpDc0F0aGNmdXBYUnhMN1cwUHBaQm53OEUwTzBBRVNYNjVaQ0JHdjZpOFRBWGhnMEpzbER5UmtmZUlnYnNHUmV2eHQxblFGZ0hNcFNpeTNWRTB3ZCIsInVzZXIiOnsiY291bnRyeSI6ImJyIiwibG9jYWxlIjoiZW5fVVMiLCJhZ2UiOnsibWluIjoyMX19LCJ1c2VyX2lkIjoiMTAwMDA1MDEyNDY2Nzg1In0'}
        request = self.get_canvas_url(data=data)
        request.user = AnonymousUser()
        mocked_method_1.return_value = {'user_id': '123456',
                                        'oauth_token': 'qwertyuiop'}
        mocked_method_2.return_value = facebook_settings.FACEBOOK_DEFAULT_SCOPE
        self.assertIsNone(self.middleware.process_request(request))
        self.assertTrue(mocked_method_1.called)



from django.contrib import admin, messages
from django.contrib.auth import models as auth_models
from django import template
from collections import defaultdict
import logging
from django_facebook.utils import get_profile
logger = logging.getLogger(__name__)


def retry_facebook_invite(modeladmin, request, queryset):
    '''
    Retries sending the invite to the users wall
    '''
    invites = list(queryset)
    user_invites = defaultdict(list)
    for invite in invites:
        user_invites[invite.user].append(invite)

    for user, invites in user_invites.items():
        profile = get_profile(user)
        graph = profile.get_offline_graph()
        if not graph:
            error_message = 'couldnt connect to the graph, user access token is %s' % profile.access_token
            messages.error(request, error_message)
            continue
        logger.info('got graph %s for user %s, retrying %s invites',
                    graph, user, len(invites))
        for invite in invites:
            invite_result = invite.resend(graph)
            message = 'User %s sent attempt to sent with id %s s6 is %s' % (
                user, invite_result.wallpost_id, not invite_result.error)
            if invite_result.error:
                message += ' got error %s' % invite_result.error_message
            messages.info(request, message)

        profile.update_invite_denormalizations()
        profile.save()


def retry_open_graph_share(modeladmin, request, queryset):
    for open_graph_share in queryset:
        open_graph_share.retry()
        messages.info(request, 'resent share %s' % open_graph_share.id)


def retry_open_graph_share_for_user(modeladmin, request, queryset):
    from django_facebook import tasks
    users = []
    for share in queryset:
        users.append(share.user)

    users = list(set(users))
    for user in users:
        tasks.retry_open_graph_shares_for_user(user)



from __future__ import unicode_literals
from django.utils.encoding import python_2_unicode_compatible
from django.conf import settings
try:
    from django.contrib.contenttypes.fields import GenericForeignKey
except ImportError:
    from django.contrib.contenttypes.generic import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.db.models.base import ModelBase
from django_facebook import model_managers, settings as facebook_settings
from open_facebook.utils import json, camel_to_underscore
from datetime import timedelta
from django_facebook.utils import compatible_datetime as datetime, \
    get_model_for_attribute, get_user_attribute, get_instance_for_attribute, \
    try_get_profile, update_user_attributes
from django_facebook.utils import get_user_model, get_profile
from open_facebook.exceptions import OAuthException
import logging
import os
logger = logging.getLogger(__name__)


def get_user_model_setting():
    from django.conf import settings
    default = 'auth.User'
    user_model_setting = getattr(settings, 'AUTH_USER_MODEL', default)
    return user_model_setting


def validate_settings():
    '''
    Checks our Facebook and Django settings and looks for common errors
    '''
    from django.conf import settings
    from django_facebook import settings as facebook_settings

    if facebook_settings.FACEBOOK_SKIP_VALIDATE:
        return

    # check for required settings
    if not facebook_settings.FACEBOOK_APP_ID:
        logger.warn('Warning FACEBOOK_APP_ID isnt specified')
    if not facebook_settings.FACEBOOK_APP_SECRET:
        logger.warn('Warning FACEBOOK_APP_SECRET isnt specified')

    # warn on things which will cause bad performance
    if facebook_settings.FACEBOOK_STORE_LIKES or facebook_settings.FACEBOOK_STORE_FRIENDS:
        if not facebook_settings.FACEBOOK_CELERY_STORE:
            msg = '''Storing friends or likes without using Celery will significantly slow down your login
Its recommended to enable FACEBOOK_CELERY_STORE or disable FACEBOOK_STORE_FRIENDS and FACEBOOK_STORE_LIKES'''
            logger.warn(msg)

    # make sure the context processors are present
    required = ['django_facebook.context_processors.facebook',
                'django.core.context_processors.request']
    context_processors = settings.TEMPLATE_CONTEXT_PROCESSORS
    for context_processor in required:
        if context_processor not in context_processors:
            logger.warn(
                'Required context processor %s wasnt found', context_processor)

    backends = settings.AUTHENTICATION_BACKENDS
    required = 'django_facebook.auth_backends.FacebookBackend'
    if required not in backends:
        logger.warn('Required auth backend %s wasnt found', required)

validate_settings()


if facebook_settings.FACEBOOK_PROFILE_IMAGE_PATH:
    PROFILE_IMAGE_PATH = settings.FACEBOOK_PROFILE_IMAGE_PATH
else:
    PROFILE_IMAGE_PATH = os.path.join('images', 'facebook_profiles/%Y/%m/%d')


class FACEBOOK_OG_STATE:

    class NOT_CONNECTED:

        '''
        The user has not connected their profile with Facebook
        '''
        pass

    class CONNECTED:

        '''
        The user has connected their profile with Facebook, but isn't
        setup for Facebook sharing
        - sharing is either disabled
        - or we have no valid access token
        '''
        pass

    class SHARING(CONNECTED):

        '''
        The user is connected to Facebook and sharing is enabled
        '''
        pass


@python_2_unicode_compatible
class BaseFacebookModel(models.Model):

    '''
    Abstract class to add to your profile or user model.
    NOTE: If you don't use this this abstract class, make sure you copy/paste
    the fields in.
    '''
    about_me = models.TextField(blank=True, null=True)
    facebook_id = models.BigIntegerField(blank=True, unique=True, null=True)
    access_token = models.TextField(
        blank=True, help_text='Facebook token for offline access', null=True)
    facebook_name = models.CharField(max_length=255, blank=True, null=True)
    facebook_profile_url = models.TextField(blank=True, null=True)
    website_url = models.TextField(blank=True, null=True)
    blog_url = models.TextField(blank=True, null=True)
    date_of_birth = models.DateField(blank=True, null=True)
    gender = models.CharField(max_length=1, choices=(
        ('m', 'Male'), ('f', 'Female')), blank=True, null=True)
    raw_data = models.TextField(blank=True, null=True)

    # the field which controls if we are sharing to facebook
    facebook_open_graph = models.NullBooleanField(
        help_text='Determines if this user want to share via open graph')

    # set to true if we require a new access token
    new_token_required = models.BooleanField(default=False,
                                             help_text='Set to true if the access token is outdated or lacks permissions')

    @property
    def open_graph_new_token_required(self):
        '''
        Shows if we need to (re)authenticate the user for open graph sharing
        '''
        reauthentication = False
        if self.facebook_open_graph and self.new_token_required:
            reauthentication = True
        elif self.facebook_open_graph is None:
            reauthentication = True
        return reauthentication

    def __str__(self):
        return self.get_user().username

    class Meta:
        abstract = True

    def refresh(self):
        '''
        Get the latest version of this object from the db
        '''
        return self.__class__.objects.get(id=self.id)

    def get_user(self):
        '''
        Since this mixin can be used both for profile and user models
        '''
        if hasattr(self, 'user'):
            user = self.user
        else:
            user = self
        return user

    def get_user_id(self):
        '''
        Since this mixin can be used both for profile and user_id models
        '''
        if hasattr(self, 'user_id'):
            user_id = self.user_id
        else:
            user_id = self.id
        return user_id

    @property
    def facebook_og_state(self):
        if not self.facebook_id:
            state = FACEBOOK_OG_STATE.NOT_CONNECTED
        elif self.access_token and self.facebook_open_graph:
            state = FACEBOOK_OG_STATE.SHARING
        else:
            state = FACEBOOK_OG_STATE.CONNECTED
        return state

    def likes(self):
        likes = FacebookLike.objects.filter(user_id=self.get_user_id())
        return likes

    def friends(self):
        friends = FacebookUser.objects.filter(user_id=self.get_user_id())
        return friends

    def disconnect_facebook(self):
        self.access_token = None
        self.new_token_required = False
        self.facebook_id = None

    def clear_access_token(self):
        self.access_token = None
        self.new_token_required = False
        self.save()

    def update_access_token(self, new_value):
        '''
        Updates the access token

        **Example**::

            # updates to 123 and sets new_token_required to False
            profile.update_access_token(123)

        :param new_value:
            The new value for access_token
        '''
        self.access_token = new_value
        self.new_token_required = False

    def extend_access_token(self):
        '''
        https://developers.facebook.com/roadmap/offline-access-removal/
        We can extend the token only once per day
        Normal short lived tokens last 1-2 hours
        Long lived tokens (given by extending) last 60 days

        The token can be extended multiple times, supposedly on every visit
        '''
        logger.info('extending access token for user %s', self.get_user())
        results = None
        if facebook_settings.FACEBOOK_CELERY_TOKEN_EXTEND:
            from django_facebook import tasks
            tasks.extend_access_token.delay(self, self.access_token)
        else:
            results = self._extend_access_token(self.access_token)
        return results

    def _extend_access_token(self, access_token):
        from open_facebook.api import FacebookAuthorization
        results = FacebookAuthorization.extend_access_token(access_token)
        access_token = results['access_token']
        old_token = self.access_token
        token_changed = access_token != old_token
        message = 'a new' if token_changed else 'the same'
        log_format = 'Facebook provided %s token, which expires at %s'
        expires_delta = timedelta(days=60)
        logger.info(log_format, message, expires_delta)
        if token_changed:
            logger.info('Saving the new access token')
            self.update_access_token(access_token)
            self.save()

        from django_facebook.signals import facebook_token_extend_finished
        facebook_token_extend_finished.send(
            sender=get_user_model(), user=self.get_user(), profile=self,
            token_changed=token_changed, old_token=old_token
        )

        return results

    def get_offline_graph(self):
        '''
        Returns a open facebook graph client based on the access token stored
        in the user's profile
        '''
        from open_facebook.api import OpenFacebook
        if self.access_token:
            graph = OpenFacebook(access_token=self.access_token)
            graph.current_user_id = self.facebook_id
            return graph

BaseFacebookProfileModel = BaseFacebookModel


class FacebookModel(BaseFacebookModel):

    '''
    the image field really destroys the subclassability of an abstract model
    you always need to customize the upload settings and storage settings

    thats why we stick it in a separate class

    override the BaseFacebookProfile if you want to change the image
    '''
    image = models.ImageField(blank=True, null=True,
                              upload_to=PROFILE_IMAGE_PATH, max_length=255)

    def profile_or_self(self):
        user_or_profile_model = get_model_for_attribute('facebook_id')
        user_model = get_user_model()
        if user_or_profile_model == user_model:
            return self
        else:
            return get_profile(self)

    class Meta:
        abstract = True


# better name for the mixin now that it can also be used for user models
FacebookProfileModel = FacebookModel


@python_2_unicode_compatible
class FacebookUser(models.Model):

    '''
    Model for storing a users friends
    '''
    # in order to be able to easily move these to an another db,
    # use a user_id and no foreign key
    user_id = models.IntegerField()
    facebook_id = models.BigIntegerField()
    name = models.TextField(blank=True, null=True)
    gender = models.CharField(choices=(
        ('F', 'female'), ('M', 'male')), blank=True, null=True, max_length=1)

    objects = model_managers.FacebookUserManager()

    class Meta:
        unique_together = ['user_id', 'facebook_id']

    def __str__(self):
        return u'Facebook user %s' % self.name


class FacebookLike(models.Model):

    '''
    Model for storing all of a users fb likes
    '''
    # in order to be able to easily move these to an another db,
    # use a user_id and no foreign key
    user_id = models.IntegerField()
    facebook_id = models.BigIntegerField()
    name = models.TextField(blank=True, null=True)
    category = models.TextField(blank=True, null=True)
    created_time = models.DateTimeField(blank=True, null=True)

    class Meta:
        unique_together = ['user_id', 'facebook_id']


class FacebookProfile(FacebookProfileModel):

    '''
    Not abstract version of the facebook profile model
    Use this by setting
    AUTH_PROFILE_MODULE = 'django_facebook.FacebookProfile'
    '''
    user = models.OneToOneField(get_user_model_setting())

if getattr(settings, 'AUTH_USER_MODEL', None) == 'django_facebook.FacebookCustomUser':
    try:
        from django.contrib.auth.models import AbstractUser, UserManager

        class FacebookCustomUser(AbstractUser, FacebookModel):

            '''
            The django 1.5 approach to adding the facebook related fields
            '''
            objects = UserManager()
            # add any customizations you like
            state = models.CharField(max_length=255, blank=True, null=True)
    except ImportError as e:
        logger.info('Couldnt setup FacebookUser, got error %s', e)


class BaseModelMetaclass(ModelBase):

    '''
    Cleaning up the table naming conventions
    '''

    def __new__(cls, name, bases, attrs):
        super_new = ModelBase.__new__(cls, name, bases, attrs)
        module_name = camel_to_underscore(name)

        app_label = super_new.__module__.split('.')[-2]
        db_table = '%s_%s' % (app_label, module_name)

        django_default = '%s_%s' % (app_label, name.lower())
        if not getattr(super_new._meta, 'proxy', False):
            db_table_is_default = django_default == super_new._meta.db_table
            # Don't overwrite when people customize the db_table
            if db_table_is_default:
                super_new._meta.db_table = db_table

        return super_new


@python_2_unicode_compatible
class BaseModel(models.Model):

    '''
    Stores the fields common to all incentive models
    '''
    __metaclass__ = BaseModelMetaclass

    def __str__(self):
        '''
        Looks at some common ORM naming standards and tries to display those before
        default to the django default
        '''
        attributes = ['name', 'title', 'slug']
        name = None
        for a in attributes:
            if hasattr(self, a):
                name = getattr(self, a)
        if not name:
            name = repr(self.__class__)
        return name

    class Meta:
        abstract = True


@python_2_unicode_compatible
class CreatedAtAbstractBase(BaseModel):

    '''
    Stores the fields common to all incentive models
    '''
    updated_at = models.DateTimeField(auto_now=True)
    created_at = models.DateTimeField(auto_now_add=True)

    # determine if we should clean this model
    auto_clean = False

    def save(self, *args, **kwargs):
        '''
        Allow for auto clean support
        '''
        if self.auto_clean:
            self.clean()
        saved = models.Model.save(self, *args, **kwargs)
        return saved

    def __str__(self):
        '''
        Looks at some common ORM naming standards and tries to display those before
        default to the django default
        '''
        attributes = ['name', 'title', 'slug']
        name = None
        for a in attributes:
            if hasattr(self, a):
                name = getattr(self, a)
        if not name:
            name = repr(self.__class__)
        return name

    def __repr__(self):
        return '<%s[%s]>' % (self.__class__.__name__, self.pk)

    class Meta:
        abstract = True


class OpenGraphShare(BaseModel):

    '''
    Object for tracking all shares to Facebook
    Used for statistics and evaluating how things are going

    I recommend running this in a task
    **Example usage**::

        from user.models import OpenGraphShare
        user = UserObject
        url = 'http://www.fashiolista.com/'
        kwargs = dict(list=url)

        share = OpenGraphShare.objects.create(
            user = user,
            action_domain='fashiolista:create',
            content_object=self,
        )
        share.set_share_dict(kwargs)
        share.save()
        result = share.send()

    **Advanced usage**::

        share.send()
        share.update(message='Hello world')
        share.remove()
        share.retry()

    Using this model has the advantage that it allows us to

    - remove open graph shares (since we store the Facebook id)

    - retry open graph shares, which is handy in case of

      - updated access tokens (retry all shares from this user in the last facebook_settings.FACEBOOK_OG_SHARE_RETRY_DAYS)

      - Facebook outages (Facebook often has minor interruptions, retry in 15m, for max facebook_settings.FACEBOOK_OG_SHARE_RETRIES)
    '''
    objects = model_managers.OpenGraphShareManager()

    user = models.ForeignKey(get_user_model_setting())

    # domain stores
    action_domain = models.CharField(max_length=255)
    facebook_user_id = models.BigIntegerField()

    # what we are sharing, dict and object
    share_dict = models.TextField(blank=True, null=True)

    content_type = models.ForeignKey(ContentType, blank=True, null=True)
    object_id = models.PositiveIntegerField(blank=True, null=True)
    content_object = GenericForeignKey('content_type', 'object_id')

    # completion data
    error_message = models.TextField(blank=True, null=True)
    last_attempt = models.DateTimeField(
        blank=True, null=True, auto_now_add=True)
    retry_count = models.IntegerField(blank=True, null=True)
    # only written if we actually succeed
    share_id = models.CharField(blank=True, null=True, max_length=255)
    completed_at = models.DateTimeField(blank=True, null=True)
    # tracking removals
    removed_at = models.DateTimeField(blank=True, null=True)

    # updated at and created at, last one needs an index
    updated_at = models.DateTimeField(auto_now=True)
    created_at = models.DateTimeField(auto_now_add=True, db_index=True)

    class Meta:
        db_table = facebook_settings.FACEBOOK_OG_SHARE_DB_TABLE

    def save(self, *args, **kwargs):
        if self.user and not self.facebook_user_id:
            profile = get_profile(self.user)
            self.facebook_user_id = get_user_attribute(
                self.user, profile, 'facebook_id')
        return BaseModel.save(self, *args, **kwargs)

    def send(self, graph=None, shared_explicitly=False):
        result = None
        # update the last attempt
        self.last_attempt = datetime.now()
        self.save()

        # see if the graph is enabled
        profile = try_get_profile(self.user)
        user_or_profile = get_instance_for_attribute(
            self.user, profile, 'access_token')
        graph = graph or user_or_profile.get_offline_graph()
        user_enabled = shared_explicitly or \
            (user_or_profile.facebook_open_graph and self.facebook_user_id)
        # start sharing
        if graph and user_enabled:
            graph_location = '%s/%s' % (
                self.facebook_user_id, self.action_domain)
            share_dict = self.get_share_dict()
            from open_facebook.exceptions import OpenFacebookException
            try:
                result = graph.set(graph_location, **share_dict)
                share_id = result.get('id')
                if not share_id:
                    error_message = 'No id in Facebook response, found %s for url %s with data %s' % (
                        result, graph_location, share_dict)
                    logger.error(error_message)
                    raise OpenFacebookException(error_message)
                self.share_id = share_id
                self.error_message = None
                self.completed_at = datetime.now()
                self.save()
            except OpenFacebookException as e:
                logger.warn(
                    'Open graph share failed, writing message %s' % str(e))
                self.error_message = repr(e)
                self.save()
                # maybe we need a new access token
                new_token_required = self.exception_requires_new_token(
                    e, graph)
                # verify that the token didnt change in the mean time
                user_or_profile = user_or_profile.__class__.objects.get(
                    id=user_or_profile.id)
                token_changed = graph.access_token != user_or_profile.access_token
                logger.info('new token required is %s and token_changed is %s',
                            new_token_required, token_changed)
                if new_token_required and not token_changed:
                    logger.info(
                        'a new token is required, setting the flag on the user or profile')
                    # time to ask the user for a new token
                    update_user_attributes(self.user, profile, dict(
                        new_token_required=True), save=True)

        elif not graph:
            self.error_message = 'no graph available'
            self.save()
        elif not user_enabled:
            self.error_message = 'user not enabled'
            self.save()

        return result

    def exception_requires_new_token(self, e, graph):
        '''
        Determines if the exceptions is something which requires us to
        ask for a new token. Examples are:

        Error validating access token: Session has expired at unix time
        1350669826. The current unix time is 1369657666.

        (#200) Requires extended permission: publish_actions (error code 200)
        '''
        new_token = False
        if isinstance(e, OAuthException):
            new_token = True

        # if we have publish actions than our token is ok
        # we get in this flow if Facebook mistakenly marks exceptions
        # as oAuthExceptions
        publish_actions = graph.has_permissions(['publish_actions'])
        if publish_actions:
            new_token = False

        return new_token

    def update(self, data, graph=None):
        '''
        Update the share with the given data
        '''
        result = None
        profile = get_profile(self.user)
        graph = graph or profile.get_offline_graph()

        # update the share dict so a retry will do the right thing
        # just in case we fail the first time
        shared = self.update_share_dict(data)
        self.save()

        # broadcast the change to facebook
        if self.share_id:
            result = graph.set(self.share_id, **shared)

        return result

    def remove(self, graph=None):
        if not self.share_id:
            raise ValueError('Can only delete shares which have an id')
        # see if the graph is enabled
        profile = get_profile(self.user)
        graph = graph or profile.get_offline_graph()
        response = None
        if graph:
            response = graph.delete(self.share_id)
            self.removed_at = datetime.now()
            self.save()
        return response

    def retry(self, graph=None, reset_retries=False):
        if self.completed_at:
            raise ValueError('You can\'t retry completed shares')

        if reset_retries:
            self.retry_count = 0
        # handle the case where self.retry_count = None
        self.retry_count = self.retry_count + 1 if self.retry_count else 1

        # actually retry now
        result = self.send(graph=graph)
        return result

    def set_share_dict(self, share_dict):
        share_dict_string = json.dumps(share_dict)
        self.share_dict = share_dict_string

    def get_share_dict(self):
        share_dict_string = self.share_dict
        share_dict = json.loads(share_dict_string)
        return share_dict

    def update_share_dict(self, share_dict):
        old_share_dict = self.get_share_dict()
        old_share_dict.update(share_dict)
        self.set_share_dict(old_share_dict)
        return old_share_dict



from django.conf import settings
from django.contrib.auth import authenticate, login
from django_facebook import settings as facebook_settings, signals
from django_facebook.connect import CONNECT_ACTIONS
from django_facebook.forms import FacebookRegistrationFormUniqueEmail
from django_facebook.utils import get_user_model, next_redirect, \
    error_next_redirect
from functools import partial
from django.contrib.auth import get_backends


class NooptRegistrationBackend(object):

    '''
    Noopt backends forms the basis of support for backends
    which handle the actual registration in the registration form
    '''

    def get_form_class(self, request):
        '''
        Returns the form class to use for registration

        :param request: the request object
        '''
        return FacebookRegistrationFormUniqueEmail

    def get_registration_template(self):
        '''
        Returns the template to use for registration
        '''
        template = facebook_settings.FACEBOOK_REGISTRATION_TEMPLATE
        return template

    def register(self, request, form=None, **kwargs):
        '''
        Implement your registration logic in this method

        :param request: the request object
        :param form: the form with the users data
        :param kwargs: additional data
        '''
        pass

    def activate(self, **kwargs):
        raise NotImplementedError

    def registration_allowed(self, request):
        return getattr(settings, 'REGISTRATION_OPEN', True)

    def post_error(self, request, additional_params=None):
        '''
        Handles the redirect after connecting
        '''
        response = error_next_redirect(
            request,
            additional_params=additional_params)
        return response

    def post_connect(self, request, user, action):
        '''
        Handles the redirect after connecting
        '''
        default_url = facebook_settings.FACEBOOK_LOGIN_DEFAULT_REDIRECT
        base_next_redirect = partial(
            next_redirect, request, default=default_url)

        if action is CONNECT_ACTIONS.LOGIN:
            response = base_next_redirect(next_key=['login_next', 'next'])
        elif action is CONNECT_ACTIONS.CONNECT:
            response = base_next_redirect(next_key=['connect_next', 'next'])
        elif action is CONNECT_ACTIONS.REGISTER:
            response = base_next_redirect(next_key=['register_next', 'next'])

        return response

    def post_activation_redirect(self, request, user):
        raise NotImplementedError


class FacebookRegistrationBackend(NooptRegistrationBackend):

    """
    A backend compatible with Django Registration
    It is extremly simple and doesn't handle things like redirects etc
    (These are already handled by Django Facebook)
    """

    def register(self, request, form=None, **kwargs):
        """
        Create and immediately log in a new user.

        """
        username, email, password = kwargs['username'], kwargs[
            'email'], kwargs['password1']
        # Create user doesn't accept additional parameters,
        new_user = get_user_model(
        ).objects.create_user(username, email, password)

        signals.user_registered.send(sender=self.__class__,
                                     user=new_user,
                                     request=request)
        authenticated_user = self.authenticate(request, username, password)
        return authenticated_user

    def authenticate(self, request, username, password):
        # authenticate() always has to be called before login(), and
        # will return the user we just created.
        authentication_details = dict(username=username, password=password)
        user = authenticate(**authentication_details)
        login(request, user)

        if user is None or not user.is_authenticated():
            backends = get_backends()
            msg_format = 'Authentication using backends %s and data %s failed'
            raise ValueError(msg_format % (backends, authentication_details))

        return user


class UserenaBackend(NooptRegistrationBackend):

    def register(self, request, form, **kwargs):
        new_user = form.save()
        return new_user

    def get_form_class(self, request):
        from userena.forms import SignupForm
        return SignupForm

    def get_registration_template(self):
        if facebook_settings.FACEBOOK_REGISTRATION_TEMPLATE == facebook_settings.default_registration_template:
            template = 'userena/signup_form.html'
        return template


class OldDjangoRegistrationBackend(NooptRegistrationBackend):

    def get_form_class(self, request):
        from registration.forms import RegistrationFormUniqueEmail
        return RegistrationFormUniqueEmail



from django.test.client import RequestFactory
from open_facebook.api import OpenFacebook, FacebookAuthorization
from django.core.handlers.base import BaseHandler


class RequestMock(RequestFactory):

    '''
    Didn't see another solution for this. Decided to read some snippets
    and modded them into the requestfactory class
    http://www.mellowmorning.com/2011/04/18/mock-django-request-for-testing/
    '''

    def request(self, **request):
        "Construct a generic request object."
        request['POST'] = dict()
        request['GET'] = dict()
        request = RequestFactory.request(self, **request)
        handler = BaseHandler()
        handler.load_middleware()
        for middleware_method in handler._request_middleware:
            if middleware_method(request):
                raise Exception("Couldn't create request mock object - "
                                "request middleware returned a response")
        return request


class MockFacebookAPI(OpenFacebook):
    mock = True

    def me(self):
        from django_facebook.test_utils.sample_user_data import user_data
        data = user_data[self.access_token]
        return data

    def my_image_url(self, size=None):
        from django_facebook.test_utils.sample_user_data import user_data
        data = user_data[self.access_token]
        image_url = data['image']
        return image_url

    def is_authenticated(self, *args, **kwargs):
        from django_facebook.test_utils.sample_user_data import user_data
        return self.access_token in user_data

    def get(self, path, **kwargs):
        if 'likes' in path:
            like = dict(
                name="Vogue Nederland", category="Media/news/publishing", id="136067283169158")
            response = dict(data=[like])
            return response
        if 'friends' or 'friend' in path:
            friend = dict(name="Aida Tavakkolie", id="172001264")
            response = dict(data=[friend])
            return response

    def set(self, path, **kwargs):
        return dict(id=123456789)

    def fql(self, query, **kwargs):
        """Runs the specified query against the Facebook FQL API.
        """
        friend = dict(name="Aida Tavakkolie", uid=172001264, gender='F')
        response = [friend]

        return response


class MockFacebookAuthorization(FacebookAuthorization):

    @classmethod
    def extend_access_token(cls, access_token):
        '''
        https://developers.facebook.com/roadmap/offline-access-removal/
        We can extend the token only once per day
        Normal short lived tokens last 1-2 hours
        Long lived tokens (given by extending) last 60 days
        '''
        response = dict(access_token=access_token, expires='123456789')
        return response



from django.contrib import auth
from django.contrib.auth import authenticate, login
from django.core.files.temp import NamedTemporaryFile
from django.core.files.uploadedfile import InMemoryUploadedFile
from django.db import transaction
from django.db.utils import IntegrityError
import json
from django_facebook import exceptions as facebook_exceptions, \
    settings as facebook_settings, signals
from django_facebook.api import get_facebook_graph
from django_facebook.utils import get_registration_backend, get_form_class, \
    get_profile_model, to_bool, get_user_model, get_instance_for,\
    get_user_attribute, try_get_profile, get_model_for_attribute,\
    get_instance_for_attribute, update_user_attributes
from random import randint
import logging
import sys
import urllib
try:
    import urllib2
except ImportError:
    import urllib.error as urllib2


logger = logging.getLogger(__name__)


class CONNECT_ACTIONS:

    class LOGIN:
        pass

    class CONNECT(LOGIN):
        pass

    class REGISTER:
        pass


def connect_user(request, access_token=None, facebook_graph=None, connect_facebook=False):
    '''
    Given a request either

    - (if authenticated) connect the user
    - login
    - register
    '''
    user = None
    graph = facebook_graph or get_facebook_graph(request, access_token)

    converter = get_instance_for('user_conversion', graph)

    assert converter.is_authenticated()
    facebook_data = converter.facebook_profile_data()
    force_registration = request.POST.get('force_registration') or \
        request.GET.get('force_registration') or \
        request.POST.get('force_registration_hard') or \
        request.GET.get('force_registration_hard')

    logger.debug('force registration is set to %s', force_registration)
    if connect_facebook and request.user.is_authenticated() and not force_registration:
        # we should only allow connect if users indicate they really want to connect
        # only when the request.CONNECT_FACEBOOK = 1
        # if this isn't present we just do a login
        action = CONNECT_ACTIONS.CONNECT
        # default behaviour is not to overwrite old data
        user = _connect_user(request, converter, overwrite=True)
    else:
        email = facebook_data.get('email', False)
        email_verified = facebook_data.get('verified', False)
        kwargs = {}
        if email and email_verified:
            kwargs = {'facebook_email': email}
        auth_user = authenticate(facebook_id=facebook_data['id'], **kwargs)
        if auth_user and not force_registration:
            action = CONNECT_ACTIONS.LOGIN

            # Has the user registered without Facebook, using the verified FB
            # email address?
            # It is after all quite common to use email addresses for usernames
            update = getattr(auth_user, 'fb_update_required', False)
            profile = try_get_profile(auth_user)
            current_facebook_id = get_user_attribute(
                auth_user, profile, 'facebook_id')
            if not current_facebook_id:
                update = True
            # login the user
            user = _login_user(request, converter, auth_user, update=update)
        else:
            action = CONNECT_ACTIONS.REGISTER
            # when force registration is active we should remove the old
            # profile
            try:
                user = _register_user(request, converter,
                                      remove_old_connections=force_registration)
            except facebook_exceptions.AlreadyRegistered as e:
                # in Multithreaded environments it's possible someone beats us to
                # the punch, in that case just login
                logger.info(
                    'parallel register encountered, slower thread is doing a login')
                auth_user = authenticate(
                    facebook_id=facebook_data['id'], **kwargs)
                if not auth_user:
                    # We don't have a valid user so raise
                    raise e
                action = CONNECT_ACTIONS.LOGIN
                user = _login_user(request, converter, auth_user, update=False)

    _update_likes_and_friends(request, user, converter)

    _update_access_token(user, graph)

    logger.info('connect finished with action %s', action)

    return action, user


def _login_user(request, facebook, authenticated_user, update=False):
    login(request, authenticated_user)

    if update:
        _connect_user(request, facebook)

    return authenticated_user


def _connect_user(request, facebook, overwrite=True):
    '''
    Update the fields on the user model and connects it to the facebook account

    '''
    if not request.user.is_authenticated():
        raise ValueError(
            'Connect user can only be used on authenticated users')
    if not facebook.is_authenticated():
        raise ValueError(
            'Facebook needs to be authenticated for connect flows')

    data = facebook.facebook_profile_data()
    facebook_id = data['id']

    # see if we already have profiles connected to this Facebook account
    old_connections = _get_old_connections(facebook_id, request.user.id)[:20]
    if old_connections and not (request.POST.get('confirm_connect') or
                                request.GET.get('confirm_connect')):
        raise facebook_exceptions.AlreadyConnectedError(list(old_connections))
    user = _update_user(request.user, facebook, overwrite=overwrite)

    return user


def _update_likes_and_friends(request, user, facebook):
    # store likes and friends if configured
    sid = transaction.savepoint()
    try:
        if facebook_settings.FACEBOOK_STORE_LIKES:
            facebook.get_and_store_likes(user)
        if facebook_settings.FACEBOOK_STORE_FRIENDS:
            facebook.get_and_store_friends(user)
        transaction.savepoint_commit(sid)
    except IntegrityError as e:
        logger.warn(u'Integrity error encountered during registration, '
                    'probably a double submission %s' % e,
                    exc_info=sys.exc_info(), extra={
                        'request': request,
                        'data': {
                            'body': unicode(e),
                        }
                    })
        transaction.savepoint_rollback(sid)


def _update_access_token(user, graph):
    '''
    Conditionally updates the access token in the database
    '''
    profile = try_get_profile(user)
    model_or_profile = get_instance_for_attribute(
        user, profile, 'access_token')
    # store the access token for later usage if the profile model supports it
    if model_or_profile:
        # update if not equal to the current token
        new_token = graph.access_token != model_or_profile.access_token
        token_message = 'a new' if new_token else 'the same'
        logger.info(
            'found %s token %s', token_message, graph.access_token[:10])
        if new_token:
            logger.info('access token changed, updating now')
            model_or_profile.update_access_token(graph.access_token)
            model_or_profile.save()
            # see if we can extend the access token
            # this runs in a task, after extending the token we fire an event
            model_or_profile.extend_access_token()


def _register_user(request, facebook, profile_callback=None,
                   remove_old_connections=False):
    '''
    Creates a new user and authenticates
    The registration form handles the registration and validation
    Other data on the user profile is updates afterwards

    if remove_old_connections = True we will disconnect old
    profiles from their facebook flow
    '''
    if not facebook.is_authenticated():
        raise ValueError(
            'Facebook needs to be authenticated for connect flows')

    # get the backend on new registration systems, or none
    # if we are on an older version
    backend = get_registration_backend()
    logger.info('running backend %s for registration', backend)

    # gets the form class specified in FACEBOOK_REGISTRATION_FORM
    form_class = get_form_class(backend, request)

    facebook_data = facebook.facebook_registration_data()

    data = request.POST.copy()
    for k, v in facebook_data.items():
        if not data.get(k):
            data[k] = v
    if remove_old_connections:
        _remove_old_connections(facebook_data['facebook_id'])

    if request.POST.get('force_registration_hard') or \
            request.GET.get('force_registration_hard'):
        data['email'] = data['email'].replace(
            '@', '+test%s@' % randint(0, 1000000000))

    form = form_class(data=data, files=request.FILES,
                      initial={'ip': request.META['REMOTE_ADDR']})

    if not form.is_valid():
        # show errors in sentry
        form_errors = form.errors
        error = facebook_exceptions.IncompleteProfileError(
            'Facebook signup incomplete')
        error.form = form
        raise error

    try:
        # for new registration systems use the backends methods of saving
        new_user = None
        if backend:
            new_user = backend.register(request,
                                        form=form, **form.cleaned_data)
        # fall back to the form approach
        if new_user is None:
            raise ValueError(
                'new_user is None, note that backward compatability for the older versions of django registration has been dropped.')
    except IntegrityError as e:
        # this happens when users click multiple times, the first request registers
        # the second one raises an error
        raise facebook_exceptions.AlreadyRegistered(e)

    # update some extra data not yet done by the form
    new_user = _update_user(new_user, facebook)

    signals.facebook_user_registered.send(sender=get_user_model(),
                                          user=new_user, facebook_data=facebook_data, request=request, converter=facebook)
    # IS this the correct way for django 1.3? seems to require the backend
    # attribute for some reason
    new_user.backend = 'django_facebook.auth_backends.FacebookBackend'
    auth.login(request, new_user)

    return new_user


def _get_old_connections(facebook_id, current_user_id=None):
    '''
    Gets other accounts connected to this facebook id, which are not
    attached to the current user
    '''
    user_or_profile_model = get_model_for_attribute('facebook_id')
    other_facebook_accounts = user_or_profile_model.objects.filter(
        facebook_id=facebook_id)
    kwargs = {}

    if current_user_id:
        # if statement since we need to support both
        user_model = get_user_model()
        if user_or_profile_model == user_model:
            kwargs['id'] = current_user_id
        else:
            kwargs['user'] = current_user_id
        other_facebook_accounts = other_facebook_accounts.exclude(**kwargs)
    return other_facebook_accounts


def _remove_old_connections(facebook_id, current_user_id=None):
    '''
    Removes the facebook id for profiles with the specified facebook id
    which arent the current user id
    '''
    other_facebook_accounts = _get_old_connections(
        facebook_id, current_user_id)
    other_facebook_accounts.update(facebook_id=None)


def _update_user(user, facebook, overwrite=True):
    '''
    Updates the user and his/her profile with the data from facebook
    '''
    # if you want to add fields to ur user model instead of the
    # profile thats fine
    # partial support (everything except raw_data and facebook_id is included)
    facebook_data = facebook.facebook_registration_data(username=False)
    facebook_fields = ['facebook_name', 'facebook_profile_url', 'gender',
                       'date_of_birth', 'about_me', 'website_url', 'first_name', 'last_name']

    profile = try_get_profile(user)
    # which attributes to update
    attributes_dict = {}

    # send the signal that we're updating
    signals.facebook_pre_update.send(sender=get_user_model(), user=user,
                                     profile=profile, facebook_data=facebook_data)

    # set the facebook id and make sure we are the only user with this id
    current_facebook_id = get_user_attribute(user, profile, 'facebook_id')
    facebook_id_changed = facebook_data['facebook_id'] != current_facebook_id
    overwrite_allowed = overwrite or not current_facebook_id

    # update the facebook id and access token
    facebook_id_overwritten = False
    if facebook_id_changed and overwrite_allowed:
        # when not overwriting we only update if there is no
        # profile.facebook_id
        logger.info('profile facebook id changed from %s to %s',
                    repr(facebook_data['facebook_id']),
                    repr(current_facebook_id))
        attributes_dict['facebook_id'] = facebook_data['facebook_id']
        facebook_id_overwritten = True

    if facebook_id_overwritten:
        _remove_old_connections(facebook_data['facebook_id'], user.id)

    # update all fields on both user and profile
    for f in facebook_fields:
        facebook_value = facebook_data.get(f, False)
        current_value = get_user_attribute(user, profile, f, None)
        if facebook_value and not current_value:
            attributes_dict[f] = facebook_value

    # write the raw data in case we missed something
    serialized_fb_data = json.dumps(facebook.facebook_profile_data())
    current_raw_data = get_user_attribute(user, profile, 'raw_data')
    if current_raw_data != serialized_fb_data:
        attributes_dict['raw_data'] = serialized_fb_data

    image_url = facebook_data['image']
    # update the image if we are allowed and have to
    if facebook_settings.FACEBOOK_STORE_LOCAL_IMAGE:
        image_field = get_user_attribute(user, profile, 'image', True)
        if not image_field:
            image_name, image_file = _update_image(
                facebook_data['facebook_id'], image_url)
            image_field.save(image_name, image_file)

    # save both models if they changed
    update_user_attributes(user, profile, attributes_dict)
    if getattr(user, '_fb_is_dirty', False):
        user.save()
    if getattr(profile, '_fb_is_dirty', False):
        profile.save()

    signals.facebook_post_update.send(sender=get_user_model(),
                                      user=user, profile=profile, facebook_data=facebook_data)

    return user


def _update_image(facebook_id, image_url):
    '''
    Updates the user profile's image to the given image url
    Unfortunately this is quite a pain to get right with Django
    Suggestions to improve this are welcome
    '''
    image_name = 'fb_image_%s.jpg' % facebook_id
    image_temp = NamedTemporaryFile()
    try:
        image_response = urllib2.urlopen(image_url)
    except AttributeError:
        image_response = urllib.request.urlopen(image_url)
    image_content = image_response.read()
    image_temp.write(image_content)
    http_message = image_response.info()
    image_size = len(image_content)
    try:
        content_type = http_message.type
    except AttributeError:
        content_type = http_message.get_content_type()
    image_file = InMemoryUploadedFile(
        file=image_temp, name=image_name, field_name='image',
        content_type=content_type, size=image_size, charset=None
    )
    image_file.seek(0)
    image_temp.flush()
    return image_name, image_file


def update_connection(request, graph):
    '''
    A special purpose view for updating the connection with an existing user
    - updates the access token (already done in get_graph)
    - sets the facebook_id if nothing is specified
    - stores friends and likes if possible
    '''
    converter = get_instance_for('user_conversion', graph)
    user = _connect_user(request, converter, overwrite=False)
    _update_likes_and_friends(request, user, converter)
    _update_access_token(user, graph)
    return user



from django.core.exceptions import ValidationError

from django_facebook import settings as facebook_settings, signals
from django_facebook.exceptions import FacebookException
from django_facebook.utils import get_user_model, mass_get_or_create, \
    cleanup_oauth_url, get_profile_model, parse_signed_request, hash_key, \
    try_get_profile, get_user_attribute
from open_facebook import exceptions as open_facebook_exceptions
from open_facebook.exceptions import OpenFacebookException
from open_facebook.utils import send_warning, validate_is_instance
import datetime
import json
import logging

try:
    from dateutil.parser import parse as parse_date
except ImportError:
    from django_facebook.utils import parse_like_datetime as parse_date


logger = logging.getLogger(__name__)


def require_persistent_graph(request, *args, **kwargs):
    '''
    Just like get_persistent graph, but instead of returning None
    raise an OpenFacebookException if we can't access facebook
    '''
    kwargs['raise_'] = True
    graph = get_persistent_graph(request, *args, **kwargs)
    if not graph:
        raise OpenFacebookException('please authenticate')
    return graph


def require_facebook_graph(request, *args, **kwargs):
    '''
    Just like get_facebook graph, but instead of returning None
    raise an OpenFacebookException if we can't access facebook
    '''
    kwargs['raise_'] = True
    graph = get_facebook_graph(request, *args, **kwargs)
    if not graph:
        raise OpenFacebookException('please authenticate')
    return graph


def get_persistent_graph(request, *args, **kwargs):
    '''
    Wraps itself around get facebook graph
    But stores the graph in the session, allowing usage across multiple
    pageviews.
    Note that Facebook session's expire at some point, you can't store this
    for permanent usage
    Atleast not without asking for the offline_access permission
    '''
    from open_facebook.api import OpenFacebook
    if not request:
        raise(ValidationError,
              'Request is required if you want to use persistent tokens')

    graph = None
    # some situations like an expired access token require us to refresh our
    # graph
    require_refresh = False
    code = request.POST.get('code', request.GET.get('code'))
    if code:
        require_refresh = True

    local_graph = getattr(request, 'facebook', None)
    if local_graph:
        # gets the graph from the local memory if available
        graph = local_graph

    if not graph:
        # search for the graph in the session
        cached_graph_dict = request.session.get('graph_dict')
        if cached_graph_dict:
            graph = OpenFacebook()
            graph.__setstate__(cached_graph_dict)
            graph._me = None

    if not graph or require_refresh:
        # gets the new graph, note this might do token conversions (slow)
        graph = get_facebook_graph(request, *args, **kwargs)
        # if it's valid replace the old cache
        if graph is not None and graph.access_token:
            request.session['graph_dict'] = graph.__getstate__()

    # add the current user id and cache the graph at the request level
    _add_current_user_id(graph, request.user)
    request.facebook = graph

    return graph


def get_facebook_graph(request=None, access_token=None, redirect_uri=None, raise_=False):
    '''
    given a request from one of these
    - js authentication flow (signed cookie)
    - facebook app authentication flow (signed cookie)
    - facebook oauth redirect (code param in url)
    - mobile authentication flow (direct access_token)
    - offline access token stored in user profile

    returns a graph object

    redirect path is the path from which you requested the token
    for some reason facebook needs exactly this uri when converting the code
    to a token
    falls back to the current page without code in the request params
    specify redirect_uri if you are not posting and recieving the code
    on the same page
    '''
    # this is not a production flow, but very handy for testing
    query_access_token = request.POST.get(
        'access_token',
        request.GET.get('access_token'))
    if not access_token and query_access_token:
        access_token = query_access_token
    # should drop query params be included in the open facebook api,
    # maybe, weird this...
    from open_facebook import OpenFacebook, FacebookAuthorization
    from django.core.cache import cache
    expires = None
    if hasattr(request, 'facebook') and request.facebook:
        graph = request.facebook
        _add_current_user_id(graph, request.user)
        return graph

    # parse the signed request if we have it
    signed_data = None
    if request:
        signed_request_string = request.POST.get(
            'signed_data',
            request.GET.get('signed_data'))
        if signed_request_string:
            logger.info('Got signed data from facebook')
            signed_data = parse_signed_request(signed_request_string)
        if signed_data:
            logger.info('We were able to parse the signed data')

    # the easy case, we have an access token in the signed data
    if signed_data and 'oauth_token' in signed_data:
        access_token = signed_data['oauth_token']

    if not access_token:
        # easy case, code is in the get
        code = request.POST.get('code', request.GET.get('code'))
        if code:
            logger.info('Got code from the request data')

        if not code:
            # signed request or cookie leading, base 64 decoding needed
            cookie_name = 'fbsr_%s' % facebook_settings.FACEBOOK_APP_ID
            cookie_data = request.COOKIES.get(cookie_name)

            if cookie_data:
                signed_request_string = cookie_data
                if signed_request_string:
                    logger.info('Got signed data from cookie')
                signed_data = parse_signed_request(signed_request_string)
                if signed_data:
                    logger.info('Parsed the cookie data')
                # the javascript api assumes a redirect uri of ''
                redirect_uri = ''

            if signed_data:
                # parsed data can fail because of signing issues
                if 'oauth_token' in signed_data:
                    logger.info('Got access_token from parsed data')
                    # we already have an active access token in the data
                    access_token = signed_data['oauth_token']
                else:
                    logger.info('Got code from parsed data')
                    # no access token, need to use this code to get one
                    code = signed_data.get('code', None)

        if not access_token:
            if code:
                cache_key = hash_key('convert_code_%s' % code)
                access_token = cache.get(cache_key)
                if not access_token:
                    # exchange the code for an access token
                    # based on the php api
                    # https://github.com/facebook/php-sdk/blob/master/src/base_facebook.php
                    # create a default for the redirect_uri
                    # when using the javascript sdk the default
                    # should be '' an empty string
                    # for other pages it should be the url
                    if not redirect_uri:
                        redirect_uri = ''

                    # we need to drop signed_data, code and state
                    redirect_uri = cleanup_oauth_url(redirect_uri)

                    try:
                        logger.info(
                            'trying to convert the code with redirect uri: %s',
                            redirect_uri)
                        # This is realy slow, that's why it's cached
                        token_response = FacebookAuthorization.convert_code(
                            code, redirect_uri=redirect_uri)
                        expires = token_response.get('expires')
                        access_token = token_response['access_token']
                        # would use cookies instead, but django's cookie setting
                        # is a bit of a mess
                        cache.set(cache_key, access_token, 60 * 60 * 2)
                    except (open_facebook_exceptions.OAuthException, open_facebook_exceptions.ParameterException) as e:
                        # this sometimes fails, but it shouldnt raise because
                        # it happens when users remove your
                        # permissions and then try to reauthenticate
                        logger.warn('Error when trying to convert code %s',
                                    str(e))
                        if raise_:
                            raise
                        else:
                            return None
            elif request.user.is_authenticated():
                # support for offline access tokens stored in the users profile
                profile = try_get_profile(request.user)
                access_token = get_user_attribute(
                    request.user, profile, 'access_token')
                if not access_token:
                    if raise_:
                        message = 'Couldnt find an access token in the request or the users profile'
                        raise open_facebook_exceptions.OAuthException(message)
                    else:
                        return None
            else:
                if raise_:
                    message = 'Couldnt find an access token in the request or cookies'
                    raise open_facebook_exceptions.OAuthException(message)
                else:
                    return None

    graph = OpenFacebook(access_token, signed_data, expires=expires)
    # add user specific identifiers
    if request:
        _add_current_user_id(graph, request.user)

    return graph


def _add_current_user_id(graph, user):
    '''
    set the current user id, convenient if you want to make sure you
    fb session and user belong together
    '''
    if graph:
        graph.current_user_id = None

        if user.is_authenticated():
            profile = try_get_profile(user)
            facebook_id = get_user_attribute(user, profile, 'facebook_id')
            if facebook_id:
                graph.current_user_id = facebook_id


class FacebookUserConverter(object):

    '''
    This conversion class helps you to convert Facebook users to Django users

    Helps with
    - extracting and prepopulating full profile data
    - invite flows
    - importing and storing likes
    '''

    def __init__(self, open_facebook):
        from open_facebook.api import OpenFacebook
        self.open_facebook = open_facebook
        validate_is_instance(open_facebook, OpenFacebook)
        self._profile = None

    def is_authenticated(self):
        return self.open_facebook.is_authenticated()

    def facebook_registration_data(self, username=True):
        '''
        Gets all registration data
        and ensures its correct input for a django registration
        '''
        facebook_profile_data = self.facebook_profile_data()
        user_data = {}
        try:
            user_data = self._convert_facebook_data(
                facebook_profile_data, username=username)
        except OpenFacebookException as e:
            self._report_broken_facebook_data(
                user_data, facebook_profile_data, e)
            raise

        return user_data

    def facebook_profile_data(self):
        '''
        Returns the facebook profile data, together with the image locations
        '''
        if self._profile is None:
            profile = self.open_facebook.me()
            profile['image'] = self.open_facebook.my_image_url('large')
            profile['image_thumb'] = self.open_facebook.my_image_url()
            self._profile = profile
        return self._profile

    @classmethod
    def _convert_facebook_data(cls, facebook_profile_data, username=True):
        '''
        Takes facebook user data and converts it to a format for
        usage with Django
        '''
        user_data = facebook_profile_data.copy()
        profile = facebook_profile_data.copy()
        website = profile.get('website')
        if website:
            user_data['website_url'] = cls._extract_url(website)

        user_data['facebook_profile_url'] = profile.get('link')
        user_data['facebook_name'] = profile.get('name')
        if len(user_data.get('email', '')) > 75:
            # no more fake email accounts for facebook
            del user_data['email']

        gender = profile.get('gender', None)

        if gender == 'male':
            user_data['gender'] = 'm'
        elif gender == 'female':
            user_data['gender'] = 'f'

        user_data['username'] = cls._retrieve_facebook_username(user_data)
        user_data['password2'], user_data['password1'] = (
            cls._generate_fake_password(),) * 2  # same as double equal

        facebook_map = dict(birthday='date_of_birth',
                            about='about_me', id='facebook_id')
        for k, v in facebook_map.items():
            user_data[v] = user_data.get(k)
        user_data['facebook_id'] = int(user_data['facebook_id'])

        if not user_data['about_me'] and user_data.get('quotes'):
            user_data['about_me'] = user_data.get('quotes')

        user_data['date_of_birth'] = cls._parse_data_of_birth(
            user_data['date_of_birth'])

        if username:
            user_data['username'] = cls._create_unique_username(
                user_data['username'])

        # make sure the first and last name are not too long
        if 'first_name' in user_data:
            user_data['first_name'] = user_data['first_name'][:30]

        if 'last_name' in user_data:
            user_data['last_name'] = user_data['last_name'][:30]

        return user_data

    @classmethod
    def _extract_url(cls, text_url_field):
        '''
        >>> from django_facebook.api import FacebookApi
        >>> url_text = 'http://www.google.com blabla'
        >>> FacebookAPI._extract_url(url_text)
        u'http://www.google.com/'

        >>> url_text = 'http://www.google.com/'
        >>> FacebookAPI._extract_url(url_text)
        u'http://www.google.com/'

        >>> url_text = 'google.com/'
        >>> FacebookAPI._extract_url(url_text)
        u'http://google.com/'

        >>> url_text = 'http://www.fahiolista.com/www.myspace.com/www.google.com'
        >>> FacebookAPI._extract_url(url_text)
        u'http://www.fahiolista.com/www.myspace.com/www.google.com'
        '''
        import re
        text_url_field = text_url_field.encode('utf8')
        seperation = re.compile('[ ,;\n\r]+')
        try:
            parts = seperation.split(text_url_field)
        except TypeError:
            parts = seperation.split(text_url_field.decode())
        for part in parts:
            from django_facebook.utils import get_url_field
            url_check = get_url_field()
            try:
                clean_url = url_check.clean(part)
                return clean_url
            except ValidationError:
                continue

    @classmethod
    def _generate_fake_password(cls):
        '''
        Returns a random fake password
        '''
        import string
        from random import choice
        size = 9
        try:
            string.letters
        except AttributeError:
            string.letters = string.ascii_letters
        password = ''.join([choice(string.letters + string.digits)
                            for i in range(size)])
        return password.lower()

    @classmethod
    def _parse_data_of_birth(cls, data_of_birth_string):
        if data_of_birth_string:
            format = '%m/%d/%Y'
            try:
                parsed_date = datetime.datetime.strptime(
                    data_of_birth_string, format)
                return parsed_date
            except ValueError:
                # Facebook sometimes provides a partial date format
                # ie 04/07 (ignore those)
                if data_of_birth_string.count('/') != 1:
                    raise

    @classmethod
    def _report_broken_facebook_data(cls, facebook_data,
                                     original_facebook_data, e):
        '''
        Sends a nice error email with the
        - facebook data
        - exception
        - stacktrace
        '''
        from pprint import pformat
        data_dump = json.dumps(original_facebook_data)
        data_dump_python = pformat(original_facebook_data)
        message_format = 'The following facebook data failed with error %s' \
                         '\n\n json %s \n\n python %s \n'
        data_tuple = (unicode(e), data_dump, data_dump_python)
        message = message_format % data_tuple
        extra_data = {
            'data_dump': data_dump,
            'data_dump_python': data_dump_python,
            'facebook_data': facebook_data,
        }
        send_warning(message, **extra_data)

    @classmethod
    def _create_unique_username(cls, base_username):
        '''
        Check the database and add numbers to the username to ensure its unique
        '''
        usernames = list(
            get_user_model().objects.filter(
                username__istartswith=base_username
            ).values_list('username', flat=True))
        usernames_lower = [u.lower() for u in usernames]
        username = str(base_username)
        i = 1
        while base_username.lower() in usernames_lower:
            base_username = username + str(i)
            i += 1
        return base_username

    @classmethod
    def _retrieve_facebook_username(cls, facebook_data):
        '''
        Search for the username in 3 places
        - public profile
        - email
        - name
        '''
        username = None

        # start by checking the public profile link (your facebook username)
        link = facebook_data.get('link')
        if link:
            username = link.split('/')[-1]
            username = cls._make_username(username)
            if username and 'profilephp' in username:
                username = None

        # try the email adress next
        if not username and 'email' in facebook_data:
            username = cls._make_username(facebook_data.get(
                'email').split('@')[0])

        # last try the name of the user
        if not username or len(username) < 4:
            username = cls._make_username(facebook_data.get('name'))

        if not username:
            raise FacebookException('couldnt figure out a username')

        return username

    @classmethod
    def _make_username(cls, username):
        '''
        Slugify the username and replace - with _ to meet username requirements
        '''
        from django.template.defaultfilters import slugify
        from unidecode import unidecode
        slugified_name = slugify(unidecode(username)).replace('-', '_')

        # consider the username min and max constraints
        slugified_name = slugified_name[:30]
        if len(username) < 4:
            slugified_name = None

        return slugified_name

    def get_and_store_likes(self, user):
        '''
        Gets and stores your facebook likes to DB
        Both the get and the store run in a async task when
        FACEBOOK_CELERY_STORE = True
        '''
        if facebook_settings.FACEBOOK_CELERY_STORE:
            from django_facebook.tasks import get_and_store_likes
            get_and_store_likes.delay(user, self)
        else:
            self._get_and_store_likes(user)

    def _get_and_store_likes(self, user):
        likes = self.get_likes()
        stored_likes = self._store_likes(user, likes)
        return stored_likes

    def get_likes(self, limit=5000):
        '''
        Parses the facebook response and returns the likes
        '''
        likes_response = self.open_facebook.get('me/likes', limit=limit)
        likes = likes_response and likes_response.get('data')
        logger.info('found %s likes', len(likes))
        return likes

    def store_likes(self, user, likes):
        '''
        Given a user and likes store these in the db
        Note this can be a heavy operation, best to do it
        in the background using celery
        '''
        if facebook_settings.FACEBOOK_CELERY_STORE:
            from django_facebook.tasks import store_likes
            store_likes.delay(user, likes)
        else:
            self._store_likes(user, likes)

    @classmethod
    def _store_likes(self, user, likes):
        current_likes = inserted_likes = None

        if likes:
            from django_facebook.models import FacebookLike
            base_queryset = FacebookLike.objects.filter(user_id=user.id)
            global_defaults = dict(user_id=user.id)
            id_field = 'facebook_id'
            default_dict = {}
            for like in likes:
                name = like.get('name')
                created_time_string = like.get('created_time')
                created_time = None
                if created_time_string:
                    created_time = parse_date(like['created_time'])
                default_dict[like['id']] = dict(
                    created_time=created_time,
                    category=like.get('category'),
                    name=name
                )
            current_likes, inserted_likes = mass_get_or_create(
                FacebookLike, base_queryset, id_field, default_dict,
                global_defaults)
            logger.debug('found %s likes and inserted %s new likes',
                         len(current_likes), len(inserted_likes))

        # fire an event, so u can do things like personalizing the users' account
        # based on the likes
        signals.facebook_post_store_likes.send(sender=get_profile_model(),
                                               user=user, likes=likes, current_likes=current_likes,
                                               inserted_likes=inserted_likes,
                                               )

        return likes

    def get_and_store_friends(self, user):
        '''
        Gets and stores your facebook friends to DB
        Both the get and the store run in a async task when
        FACEBOOK_CELERY_STORE = True
        '''
        if facebook_settings.FACEBOOK_CELERY_STORE:
            from django_facebook.tasks import get_and_store_friends
            get_and_store_friends.delay(user, self)
        else:
            self._get_and_store_friends(user)

    def _get_and_store_friends(self, user):
        '''
        Getting the friends via fb and storing them
        '''
        friends = self.get_friends()
        stored_friends = self._store_friends(user, friends)
        return stored_friends

    def get_friends(self, limit=5000):
        '''
        Connects to the facebook api and gets the users friends
        '''
        friends = getattr(self, '_friends', None)
        if friends is None:
            friends_response = self.open_facebook.get('me/friends', limit=limit, fields='gender,name')

            friends = []
            for response_dict in friends_response.get('data'):
                response_dict['id'] = response_dict['id']
                friends.append(response_dict)

        logger.info('found %s friends', len(friends))

        return friends

    def store_friends(self, user, friends):
        '''
        Stores the given friends locally for this user
        Quite slow, better do this using celery on a secondary db
        '''
        if facebook_settings.FACEBOOK_CELERY_STORE:
            from django_facebook.tasks import store_friends
            store_friends.delay(user, friends)
        else:
            self._store_friends(user, friends)

    @classmethod
    def _store_friends(self, user, friends):
        from django_facebook.models import FacebookUser
        current_friends = inserted_friends = None

        # store the users for later retrieval
        if friends:
            # see which ids this user already stored
            base_queryset = FacebookUser.objects.filter(user_id=user.id)
            # if none if your friend have a gender clean the old data
            genders = FacebookUser.objects.filter(
                user_id=user.id, gender__in=('M', 'F')).count()
            if not genders:
                FacebookUser.objects.filter(user_id=user.id).delete()

            global_defaults = dict(user_id=user.id)
            default_dict = {}
            gender_map = dict(female='F', male='M')
            gender_map['male (hidden)'] = 'M'
            gender_map['female (hidden)'] = 'F'
            for f in friends:
                name = f.get('name')
                gender = None
                if f.get('sex'):
                    gender = gender_map[f.get('sex')]
                default_dict[str(f['id'])] = dict(name=name, gender=gender)
            id_field = 'facebook_id'

            current_friends, inserted_friends = mass_get_or_create(
                FacebookUser, base_queryset, id_field, default_dict,
                global_defaults)
            logger.debug('found %s friends and inserted %s new ones',
                         len(current_friends), len(inserted_friends))

        # fire an event, so u can do things like personalizing suggested users
        # to follow
        signals.facebook_post_store_friends.send(sender=get_profile_model(),
                                                 user=user, friends=friends, current_friends=current_friends,
                                                 inserted_friends=inserted_friends,
                                                 )

        return friends

    def registered_friends(self, user):
        '''
        Returns all profile models which are already registered on your site
        and a list of friends which are not on your site
        '''
        profile_class = get_profile_model()
        friends = self.get_friends(limit=1000)

        if friends:
            friend_ids = [f['id'] for f in friends]
            friend_objects = profile_class.objects.filter(
                facebook_id__in=friend_ids).select_related('user')
            registered_ids = [f.facebook_id for f in friend_objects]
            new_friends = [f for f in friends if f['id'] not in registered_ids]
        else:
            new_friends = []
            friend_objects = profile_class.objects.none()

        return friend_objects, new_friends



__author__ = 'Thierry Schellenbach'
__copyright__ = 'Copyright 2010, Thierry Schellenbach'
__credits__ = ['Thierry Schellenbach, mellowmorning.com, @tschellenbach']


__license__ = 'BSD'
__version__ = '6.0.6'
__maintainer__ = 'Thierry Schellenbach'
__email__ = 'thierryschellenbach@gmail.com'
__status__ = 'Production'

'''
Some links which help me with publishing this code

rest editor
http://rst.ninjs.org/

updating pypi
python setup.py sdist upload
http://pypi.python.org/pypi

setting up pip for editing
http://www.pip-installer.org/en/latest/index.html
pip install -e ./
'''



import json


def json_to_python(json_string):
    '''
    I receive error mails encoded as json, this function converts the data into
    easy usable python dicts and pprint them
    '''

    python_dict = json.loads(json_string)
    return python_dict

if __name__ == '__main__':
    json_string = r'''{"website": "http://fernandaferrervazquez.blogspot.com/\r\nhttp://twitter.com/fferrervazquez\r\nhttp://comunidad.redfashion.es/profile/fernandaferrervazquez\r\nhttp://www.facebook.com/group.php?gid3D40257259997&ref3Dts\r\nhttp://fernandaferrervazquez.spaces.live.com/blog/cns!EDCBAC31EE9D9A0C!326.trak\r\nhttp://www.linkedin.com/myprofile?trk3Dhb_pro\r\nhttp://www.youtube.com/account#profile\r\nhttp://www.flickr.com/\r\n Mi galer\u00eda\r\nhttp://www.flickr.com/photos/wwwfernandaferrervazquez-showroomrecoletacom/ \r\n\r\nhttp://www.facebook.com/pages/Buenos-Aires-Argentina/Fernanda-F-Showroom-Recoleta/200218353804?ref3Dts\r\nhttp://fernandaferrervazquez.wordpress.com/wp-admin/", "last_name": "Ferrer Vazquez", "locale": "es_LA", "hometown": {"id": "109238842431095", "name": "La Coru\u00f1a, Galicia, Spain"}, "image": "https://graph.facebook.com/me/picture?type3Dlarge&access_token3D100314226679773|2.1ZkFz1Cusu5RlY1xGft_Pg__.86400.1303534800.4-1157872766|ZgsRrRYwp2pqHEHtncqS5-rBiSg"
    , "timezone": -3, "education": [{"school": {"id": "115805661780893", "name": "uba arquitectura dise\u00f1o y urbanismo"}, "classes": [{"description": "direccion de arte", "id": "188729917815843", "name": "AAAP"}, {"with": [{"id": "546455670", "name": "Gustavo Lento"}], "id": "200142039998266", "name": "Indumentaria"}, {"description": "dise\u00f1o ", "with": [{"id": "1285841477", "name": "La Maison Madrid"}], "id": "102056769875267", "name": "figurin"}, {"id": "194463713907701", "name": "indumentaria -figurin-seriado"}, {"with": [{"id": "704003068", "name": "Mariano Toledo"}], "id": "180792848632395", "name": "indumentaria-figurin -seriado.etc"}], "type": "College", "concentration": [{"id": "176664472378719", "name": "indumentaria y dise\u00f1o textil"}]}, {"school": {"id": "107827585907242", "name": "Miami International University of Art & Design"}, "classes": [{"id": "116042535136769", "name": "whadrove asesory"}], "type": "Graduate School", "concentration": [{"id": "146715355391868", "name": "producer-  stylist -  asesora de imagen -personal-shooper -whadrove asesory"}]}, {"school": {"id": "105548239478812", "name": "New York Institute of Technology"}, "classes": [{"id": "199608053396849", "name": "Reporter, Producer and Director"}, {"id": "209828275699578", "name": "personal shooper"}, {"id": "208913322452336", "name": "Designer-stylisti"}], "type": "College", "concentration": [{"id": "184808144897752", "name": "whadrobing"}]}, {"school": {"id": "106431826058746", "name": "Colegio Nuestra Se\u00f1ora de la Misericordia"}, "type": "High School"}, {"school": {"id": "106011862771707", "name": "Mary E. Graham"}, "type": "High School"}, {"school": {"id": "106011862771707", "name": "Mary E. Graham"}, "type": "High School"}], "id": "1157872766", "first_name": "Fernanda", "verified": true, "languages": [{"id": "111021838921932", "name": "Espa\u00f1ol"}, {"id": "110867825605119", "name": "Ingles"}, {"id": "108083115891989", "name": "Portugu\u00e9s"}, {"id": "113051562042989", "name": "Italiano"}], "location": {"id": "106423786059675", "name": "Buenos Aires, Argentina"}, "image_thumb": "https://graph.facebook.com/me/picture?access_token3D100314226679773|2.1ZkFz1Cusu5RlY1xGft_Pg__.86400.1303534800.4-1157872766|ZgsRrRYwp2pqHEHtncqS5-rBiSg", "email": "fer-fer-666@hotmail.com", "username": "FernandaFerrerVazquez", "bio": "ATELIER : TODOS LOS DIAS DE 14 A 22 HS C CITA PREVIA. MIX DE ESTETICAS - ECLECTICA - PRENDAS UNICAS-HAND MADE- SEASSON LESS RETROMODERNIDAD -CUSTOMIZADO !!!\r\n\r\nMe encanta todo el proceso que lleva lograr un objeto que sea unico....personalizarlo de acuerdo a cada uno...a cada persona....a su forma de ser.......elegir los materiales, los avios, las telas...siendo la premisa DIFERENCIARSE!!!!\r\n Atemporal..Retro modernidad...Eclecticismo!!!", "birthday": "04/09/1973", "link": "http://www.facebook.com/FernandaFerrerVazquez", "name": "Fernanda Ferrer Vazquez", "gender": "female", "work": [{"description": "ATELIER : un mix de esteticas que la hacen eclectica, multicultural, excentrica, customizada, artesanal, unica , exclusiva, lujosa, eco-reciclada,con bordados a mano y texturas..especial atencion a los peque\u00f1os detalles!!!! \nSeason less \nHand made ", "employer": {"id": "113960871952651", "name": "fernanda ferrer vazquez"}, "location": {"id": "106423786059675", "name": "Buenos Aires, Argentina"}, "position": {"id": "145270362165639", "name": "Dise\u00f1adora"}, "start_date": "2008-12", "projects": [{"description": "produccion-estilismo-fotografia-make up-pelo-asesoramiento de imagen-whadrobing-perso\u00f1an shooper-rent an outfit!!", "id": "200218353804", "name": "Fernanda F -Showroom Recoleta"}]}, {"employer": {"id": "198034683561689", "name": "el atelier de la isla"}}], "updated_time": "2011-04-22T04:28:27+0000"}'''
    json_to_python(json_string)



from django.utils import six
from functools import wraps
try:
    # using compatible_datetime instead of datetime only
    # not to override the original datetime package
    from django.utils import timezone as compatible_datetime
except ImportError:
    from datetime import datetime as compatible_datetime
from datetime import datetime
from django.apps import apps
from django.http import QueryDict, HttpResponse, HttpResponseRedirect
from django.conf import settings
import django.contrib.auth
from django.db import models, transaction
import logging
import django
import re
from django_facebook import settings as facebook_settings
from django.utils.encoding import iri_to_uri
from django.template.loader import render_to_string
import gc

django_version = django.VERSION

logger = logging.getLogger(__name__)


class NOTHING:
    pass

'''
TODO, write an abstraction class for reading and writing users/profile models
'''


def get_profile_model():
    '''
    Get the profile model if present otherwise return None
    '''
    model = None
    profile_string = getattr(settings, 'AUTH_PROFILE_MODULE', None)
    if profile_string:
        app_label, model_label = profile_string.split('.')
        model = apps.get_model(app_label, model_label)
    return model


def get_user_model():
    '''
    For Django < 1.5 backward compatibility
    '''
    if hasattr(django.contrib.auth, 'get_user_model'):
        return django.contrib.auth.get_user_model()
    else:
        return django.contrib.auth.models.User


def get_model_for_attribute(attribute):
    if is_profile_attribute(attribute):
        model = get_profile_model()
    else:
        model = get_user_model()
    return model


def is_profile_attribute(attribute):
    profile_model = get_profile_model()
    profile_fields = []
    if profile_model:
        profile_fields = [f.name for f in profile_model._meta.fields]
    return attribute in profile_fields


def is_user_attribute(attribute):
    user_model = get_user_model()
    user_fields = [f.name for f in user_model._meta.fields]
    return attribute in user_fields


def get_instance_for_attribute(user, profile, attribute):
    profile_fields = []
    if profile:
        profile_fields = [f.name for f in profile._meta.fields]
    user_fields = [f.name for f in user._meta.fields]
    is_profile_field = lambda f: f in profile_fields and hasattr(profile, f)
    is_user_field = lambda f: f in user_fields and hasattr(user, f)

    instance = None
    if is_profile_field(attribute):
        instance = profile
    elif is_user_field(attribute):
        instance = user
    return instance


def get_user_attribute(user, profile, attribute, default=NOTHING):
    profile_fields = []
    if profile:
        profile_fields = [f.name for f in profile._meta.fields]
    user_fields = [f.name for f in user._meta.fields]
    is_profile_field = lambda f: f in profile_fields and hasattr(profile, f)
    is_user_field = lambda f: f in user_fields and hasattr(user, f)

    if is_profile_field(attribute):
        value = getattr(profile, attribute)
    elif is_user_field(attribute):
        value = getattr(user, attribute)
    elif default is not NOTHING:
        value = default
    else:
        raise AttributeError(
            'user or profile didnt have attribute %s' % attribute)

    return value


def update_user_attributes(user, profile, attributes_dict, save=False):
    '''
    Write the attributes either to the user or profile instance
    '''
    profile_fields = []
    if profile:
        profile_fields = [f.name for f in profile._meta.fields]
    user_fields = [f.name for f in user._meta.fields]

    is_profile_field = lambda f: f in profile_fields and hasattr(profile, f)
    is_user_field = lambda f: f in user_fields and hasattr(user, f)

    for f, value in attributes_dict.items():
        if is_profile_field(f):
            setattr(profile, f, value)
            profile._fb_is_dirty = True
        elif is_user_field(f):
            setattr(user, f, value)
            user._fb_is_dirty = True
        else:
            logger.info('skipping update of field %s', f)

    if save:
        if getattr(user, '_fb_is_dirty', False):
            user.save()
        if profile and getattr(profile, '_fb_is_dirty', False):
            profile.save()


def try_get_profile(user):
    try:
        p = get_profile(user)
    except:
        p = None
    return p


def hash_key(key):
    import hashlib
    key = key.encode('utf-8')
    return hashlib.md5(key).hexdigest()


def parse_signed_request(signed_request_string):
    '''
    Just here for your convenience, actual logic is in the
    FacebookAuthorization class
    '''
    from open_facebook.api import FacebookAuthorization
    signed_request = FacebookAuthorization.parse_signed_data(
        signed_request_string)
    return signed_request


def get_url_field():
    '''
    This should be compatible with both django 1.3, 1.4 and 1.5
    In 1.5 the verify_exists argument is removed and always False
    '''
    from django.forms import URLField
    field = URLField()
    try:
        field = URLField(verify_exists=False)
    except TypeError as e:
        pass
    return field


def clear_persistent_graph_cache(request):
    '''
    Clears the caches for the graph cache
    '''
    request.facebook = None
    request.session.delete('graph')
    if request.user.is_authenticated():
        profile = get_profile(request.user)
        profile.clear_access_token()


def has_permissions(graph, scope_list):
    from open_facebook import exceptions as open_facebook_exceptions
    permissions_granted = False
    try:
        if graph:
            permissions_granted = graph.has_permissions(scope_list)
    except open_facebook_exceptions.OAuthException as e:
        pass
    return permissions_granted


def queryset_iterator(queryset, chunksize=1000, getfunc=getattr):
    '''''
    Iterate over a Django Queryset ordered by the primary key

    This method loads a maximum of chunksize (default: 1000) rows in it's
    memory at the same time while django normally would load all rows in it's
    memory. Using the iterator() method only causes it to not preload all the
    classes.

    Note that the implementation of the iterator does not support ordered query sets.
    '''
    pk = 0

    try:
        '''In the case of an empty list, return'''
        last_pk = getfunc(queryset.order_by('-pk')[0], 'pk')
    except IndexError:
        return

    queryset = queryset.order_by('pk')
    while pk < last_pk:
        for row in queryset.filter(pk__gt=pk)[:chunksize].iterator():
            pk = getfunc(row, 'pk')
            yield row
        gc.collect()


def get_oauth_url(scope, redirect_uri, extra_params=None):
    '''
    Returns the oAuth URL for the given scope and redirect_uri
    '''
    scope = parse_scope(scope)
    query_dict = QueryDict('', True)
    query_dict['scope'] = ','.join(scope)
    query_dict['client_id'] = facebook_settings.FACEBOOK_APP_ID

    query_dict['redirect_uri'] = redirect_uri
    oauth_url = 'https://www.facebook.com/dialog/oauth?'
    oauth_url += query_dict.urlencode()
    return oauth_url


class ScriptRedirect(HttpResponse):

    '''
    Redirect for Facebook Canvas pages
    '''

    def __init__(self, redirect_to, show_body=True):
        self.redirect_to = redirect_to
        self.location = iri_to_uri(redirect_to)

        context = dict(location=self.location,
                       show_body=show_body)
        js_redirect = render_to_string(
            'django_facebook/canvas_redirect.html', context)

        super(ScriptRedirect, self).__init__(js_redirect)


def response_redirect(redirect_url, script_redirect=False):
    '''
    Abstract away canvas redirects
    '''
    if script_redirect:
        return ScriptRedirect(redirect_url)

    return HttpResponseRedirect(redirect_url)


def error_next_redirect(request, default='/', additional_params=None, next_key=None, redirect_url=None, canvas=False):
    '''
    Short cut for an error next redirect
    '''
    if not next_key:
        next_key = ['error_next', 'next']

    redirect = next_redirect(
        request, default, additional_params, next_key, redirect_url, canvas)
    return redirect


def next_redirect(request, default='/', additional_params=None,
                  next_key='next', redirect_url=None, canvas=False):
    if facebook_settings.FACEBOOK_DEBUG_REDIRECTS:
        return HttpResponse(
            '<html><head></head><body><div>Debugging</div></body></html>')
    from django.http import HttpResponseRedirect
    if not isinstance(next_key, (list, tuple)):
        next_key = [next_key]

    # get the redirect url
    if not redirect_url:
        for key in next_key:
            redirect_url = request.POST.get(key, request.GET.get(key))
            if redirect_url:
                break
        if not redirect_url:
            redirect_url = default

    if additional_params:
        query_params = QueryDict('', True)
        query_params.update(additional_params)
        seperator = '&' if '?' in redirect_url else '?'
        redirect_url += seperator + query_params.urlencode()

    if canvas:
        return ScriptRedirect(redirect_url)

    return HttpResponseRedirect(redirect_url)


@transaction.atomic
def mass_get_or_create(model_class, base_queryset, id_field, default_dict,
                       global_defaults):
    '''
    Updates the data by inserting all not found records
    Doesnt delete records if not in the new data

    example usage
    >>> model_class = ListItem #the class for which you are doing the insert
    >>> base_query_set = ListItem.objects.filter(user=request.user, list=1) #query for retrieving currently stored items
    >>> id_field = 'user_id' #the id field on which to check
    >>> default_dict = {'12': dict(comment='my_new_item'), '13': dict(comment='super')} #list of default values for inserts
    >>> global_defaults = dict(user=request.user, list_id=1) #global defaults
    '''
    current_instances = list(base_queryset)
    current_ids = set(
        [unicode(getattr(c, id_field)) for c in current_instances])
    given_ids = map(unicode, default_dict.keys())
    # both ends of the comparison are in unicode ensuring the not in works
    new_ids = [g for g in given_ids if g not in current_ids]
    prepared_models = []
    for new_id in new_ids:
        defaults = default_dict[new_id]
        defaults[id_field] = new_id
        defaults.update(global_defaults)
        model_instance = model_class(
            **defaults
        )
        prepared_models.append(model_instance)
    # efficiently create these objects all at once
    # django 1.4 only
    if hasattr(model_class.objects, 'bulk_create'):
        model_class.objects.bulk_create(prepared_models)
    else:
        [m.save() for m in prepared_models]
    inserted_model_instances = prepared_models
    # returns a list of existing and new items
    return current_instances, inserted_model_instances


def get_form_class(backend, request):
    '''
    Will use registration form in the following order:
    1. User configured RegistrationForm
    2. backend.get_form_class(request) from django-registration 0.8
    3. RegistrationFormUniqueEmail from django-registration < 0.8
    '''
    form_class = None

    # try the setting
    form_class_string = facebook_settings.FACEBOOK_REGISTRATION_FORM
    if form_class_string:
        try:
            form_class = get_class_from_string(form_class_string, None)
        except ImportError:
            # Shouldn't this fail -- if you set it, it must be correct?
            pass

    if not form_class:
        backend = backend or get_registration_backend()
        if backend:
            form_class = backend.get_form_class(request)

    assert form_class, 'we couldnt find a form class, so we cant go on like this'

    return form_class


def get_registration_backend():
    '''
    Ensures compatability with the new and old version of django registration
    '''
    backend = None
    backend_class = None

    registration_backend_string = getattr(
        facebook_settings, 'FACEBOOK_REGISTRATION_BACKEND', None)
    if registration_backend_string:
        backend_class = get_class_from_string(registration_backend_string)

    # instantiate
    if backend_class:
        backend = backend_class()

    return backend


def get_django_registration_version():
    '''
    Returns new, old or None depending on the version of django registration
    Old works with forms
    New works with backends
    '''
    try:
        # support for the newer implementation
        from registration.backends import get_backend
        version = 'new'
    except ImportError:
        version = 'old'

    try:
        import registration
    except ImportError as e:
        version = None

    return version


def parse_scope(scope):
    '''
    Turns
    'email,user_about_me'
    or
    ('email','user_about_me')
    into a nice consistent
    ['email','user_about_me']
    '''
    assert scope, 'scope is required'
    if isinstance(scope, six.string_types):
        scope_list = scope.split(',')
    elif isinstance(scope, (list, tuple)):
        scope_list = list(scope)
    else:
        raise ValueError('unrecognized type for scope %r' % scope)

    return scope_list


def simplify_class_decorator(class_decorator):
    '''
    Makes the decorator syntax uniform
    Regardless if you call the decorator like

    **Decorator examples**::
        @decorator
        or
        @decorator()
        or
        @decorator(staff=True)

    Complexity, Python's class based decorators are weird to say the least:
    http://www.artima.com/weblogs/viewpost.jsp?thread=240845

    This function makes sure that your decorator class always gets called with

    **Methods called**::

        __init__(fn, *option_args, *option_kwargs)
        __call__()
            return a function which accepts the *args and *kwargs intended
            for fn
    '''
    # this makes sure the resulting decorator shows up as
    # function FacebookRequired instead of outer
    @wraps(class_decorator)
    def outer(fn=None, *decorator_args, **decorator_kwargs):
        # wraps isn't needed, the decorator should do the wrapping :)
        # @wraps(fn, assigned=available_attrs(fn))
        def actual_decorator(fn):
            instance = class_decorator(fn, *decorator_args, **decorator_kwargs)
            _wrapped_view = instance.__call__()
            return _wrapped_view

        if fn is not None:
            wrapped_view = actual_decorator(fn)
        else:
            wrapped_view = actual_decorator

        return wrapped_view
    return outer


def to_int(input, default=0, exception=(ValueError, TypeError), regexp=None):
    '''Convert the given input to an integer or return default

    When trying to convert the exceptions given in the exception parameter
    are automatically catched and the default will be returned.

    The regexp parameter allows for a regular expression to find the digits
    in a string.
    When True it will automatically match any digit in the string.
    When a (regexp) object (has a search method) is given, that will be used.
    WHen a string is given, re.compile will be run over it first

    The last group of the regexp will be used as value
    '''
    if regexp is True:
        regexp = re.compile('(\d+)')
    elif isinstance(regexp, six.string_types):
        regexp = re.compile(regexp)
    elif hasattr(regexp, 'search'):
        pass
    elif regexp is not None:
        raise(TypeError, 'unknown argument for regexp parameter')

    try:
        if regexp:
            match = regexp.search(input)
            if match:
                input = match.groups()[-1]
        return int(input)
    except exception:
        return default


def to_bool(input, default=False):
    '''
    Take a request value and turn it into a bool
    Never raises errors
    '''
    if input is None:
        value = default
    else:
        int_value = to_int(input, default=None)
        if int_value is None:
            value = default
        else:
            value = bool(int_value)
    return value


def remove_query_param(url, key):
    p = re.compile('%s=[^=&]*&' % key, re.VERBOSE)
    url = p.sub('', url)
    p = re.compile('%s=[^=&]*' % key, re.VERBOSE)
    url = p.sub('', url)
    return url


def replace_query_param(url, key, value):
    p = re.compile('%s=[^=&]*' % key, re.VERBOSE)
    return p.sub('%s=%s' % (key, value), url)


DROP_QUERY_PARAMS = ['code', 'signed_request', 'state']


def cleanup_oauth_url(redirect_uri):
    '''
    We have to maintain order with respect to the
    queryparams which is a bit of a pain
    TODO: Very hacky will subclass QueryDict to SortedQueryDict at some point
    And use a decent sort function
    '''
    if '?' in redirect_uri:
        redirect_base, redirect_query = redirect_uri.split('?', 1)
        query_dict_items = QueryDict(redirect_query).items()
    else:
        query_dict_items = QueryDict('', True)

    # filtered_query_items = [(k, v) for k, v in query_dict_items
    #                         if k.lower() not in DROP_QUERY_PARAMS]
    # new_query_dict = QueryDict('', True)
    # new_query_dict.update(dict(filtered_query_items))

    excluded_query_items = [(k, v) for k, v in query_dict_items
                            if k.lower() in DROP_QUERY_PARAMS]
    for k, v in excluded_query_items:
        redirect_uri = remove_query_param(redirect_uri, k)

    redirect_uri = redirect_uri.strip('?')
    redirect_uri = redirect_uri.strip('&')

    return redirect_uri


def replication_safe(f):
    '''
    Usually views which do a POST will require the next page to be
    read from the master database. (To prevent issues with replication lag).

    However certain views like login do not have this issue.
    They do a post, but don't modify data which you'll show on subsequent pages.

    This decorators marks these views as safe.
    This ensures requests on the next page are allowed to use the slave db
    '''
    from functools import wraps

    @wraps(f)
    def wrapper(request, *args, **kwargs):
        request.replication_safe = True
        response = f(request, *args, **kwargs)
        return response

    return wrapper


def get_class_from_string(path, default=None):
    """
    Return the class specified by the string.

    IE: django.contrib.auth.models.User
    Will return the user class or cause an ImportError
    """
    try:
        from importlib import import_module
    except ImportError:
        from django.utils.importlib import import_module
    i = path.rfind('.')
    module, attr = path[:i], path[i + 1:]
    mod = import_module(module)
    try:
        return getattr(mod, attr)
    except AttributeError:
        if default:
            return default
        else:
            raise ImportError(
                'Cannot import name {} (from {})'.format(attr, mod))


def parse_like_datetime(dt):
    return datetime.strptime(dt, "%Y-%m-%dT%H:%M:%S+0000")


def get_default_mapping():
    from django_facebook.api import FacebookUserConverter
    DEFAULT_FACEBOOK_CLASS_MAPPING = {
        'user_conversion': FacebookUserConverter
    }
    return DEFAULT_FACEBOOK_CLASS_MAPPING


def get_class_mapping():
    mapping = facebook_settings.FACEBOOK_CLASS_MAPPING
    if mapping is None:
        mapping = get_default_mapping()
    return mapping


def get_class_for(purpose):
    '''
    Usage:
    conversion_class = get_class_for('user_conversion')
    '''
    mapping = get_class_mapping()
    class_ = mapping[purpose]
    if isinstance(class_, six.string_types):
        class_ = get_class_from_string(class_)
    return class_


def get_instance_for(purpose, *args, **kwargs):
    '''
    **Usage**::

        conversion_instance = get_instance_for(
            'facebook_user_conversion', user=user)
    '''
    class_ = get_class_for(purpose)
    instance = class_(*args, **kwargs)
    return instance


def get_migration_data():
    '''
    Support for Django custom user models
    See this blog post for inspiration

    http://kevindias.com/writing/django-custom-user-models-south-and-reusable-apps/
    https://github.com/stephenmcd/mezzanine/blob/master/mezzanine/core/migrations/0005_auto__chg_field_sitepermission_user__del_unique_sitepermission_user.py
    '''

    try:
        from django.contrib.auth import get_user_model
    except ImportError:  # django < 1.5
        from django.contrib.auth.models import User
    else:
        User = get_user_model()

    user_orm_label = '%s.%s' % (User._meta.app_label, User._meta.object_name)
    user_model_label = '%s.%s' % (User._meta.app_label, User._meta.module_name)
    return User, user_orm_label, user_model_label


def get_profile(user):
    '''
    Get profile
    '''
    if django_version >= (1, 7, 0):
        profile = get_profile_from_user(user)
    else:
        profile = user.get_profile()
    return profile


def get_profile_from_user(user):
    '''
    Tries to get the profile according to the
    class configured on AUTH_PROFILE_MODULE
    '''
    for field in user._meta.get_fields():
        try:
            if hasattr(user, field.name):
                attribute = getattr(user, field.name)
                if get_profile_model() == type(attribute):
                    return attribute
        except Exception:
            logger.exception("Error getting profile attribute from user.")

    logger.info("Could not find profile attribute.")
    return None



from django.conf import settings
import logging

logger = logging.getLogger(__name__)

# : Your facebook app id
FACEBOOK_APP_ID = getattr(settings, 'FACEBOOK_APP_ID', None)
# : Your facebook app secret
FACEBOOK_APP_SECRET = getattr(settings, 'FACEBOOK_APP_SECRET', None)
# : The default scope we should use, note that registration will break without email
FACEBOOK_DEFAULT_SCOPE = getattr(settings, 'FACEBOOK_DEFAULT_SCOPE', [
    'email', 'user_about_me', 'user_birthday', 'user_website'])

# : If we should store likes
FACEBOOK_STORE_LIKES = getattr(settings, 'FACEBOOK_STORE_LIKES', False)
# : If we should store friends
FACEBOOK_STORE_FRIENDS = getattr(settings, 'FACEBOOK_STORE_FRIENDS', False)

# : If we should be using celery to store friends and likes (recommended)
FACEBOOK_CELERY_STORE = getattr(settings, 'FACEBOOK_CELERY_STORE', False)
# : Use celery for updating tokens, recommended since it's quite slow
FACEBOOK_CELERY_TOKEN_EXTEND = getattr(
    settings, 'FACEBOOK_CELERY_TOKEN_EXTEND', False)

default_registration_backend = 'django_facebook.registration_backends.FacebookRegistrationBackend'
# : Allows you to overwrite the registration backend
# : Specify a full path to a class (defaults to django_facebook.registration_backends.FacebookRegistrationBackend)
FACEBOOK_REGISTRATION_BACKEND = getattr(
    settings, 'FACEBOOK_REGISTRATION_BACKEND', default_registration_backend)

# Absolute canvas page url as per facebook standard
FACEBOOK_CANVAS_PAGE = getattr(settings, 'FACEBOOK_CANVAS_PAGE',
                               'http://apps.facebook.com/django_facebook_test/')

# Disable this setting if you don't want to store a local image
FACEBOOK_STORE_LOCAL_IMAGE = getattr(
    settings, 'FACEBOOK_STORE_LOCAL_IMAGE', True)

# Track all raw data coming in from FB
FACEBOOK_TRACK_RAW_DATA = getattr(settings, 'FACEBOOK_TRACK_RAW_DATA', False)


FACEBOOK_DEBUG_REDIRECTS = getattr(settings, 'FACEBOOK_DEBUG_REDIRECTS', False)

# READ only mode, convenient when doing load testing etc.
FACEBOOK_READ_ONLY = getattr(settings, 'FACEBOOK_READ_ONLY', False)

# Allow custom registration template
default_registration_template = [
    'django_facebook/registration.html', 'registration/registration_form.html']
FACEBOOK_REGISTRATION_TEMPLATE = getattr(settings,
                                         'FACEBOOK_REGISTRATION_TEMPLATE', default_registration_template)

# Allow custom signup form
FACEBOOK_REGISTRATION_FORM = getattr(settings,
                                     'FACEBOOK_REGISTRATION_FORM', None)


# Fall back redirect location when no other location was found
FACEBOOK_LOGIN_DEFAULT_REDIRECT = getattr(
    settings, 'FACEBOOK_LOGIN_DEFAULT_REDIRECT', '/')

# Force profile update every login
FACEBOOK_FORCE_PROFILE_UPDATE_ON_LOGIN = getattr(
    settings, 'FACEBOOK_FORCE_PROFILE_UPDATE_ON_LOGIN', False)


# Retry an open graph share 6 times (once every 15 minutes)
FACEBOOK_OG_SHARE_RETRIES = getattr(settings, 'FACEBOOK_OG_SHARE_RETRIES', 6)
# Retry a failed open graph share (when we have an updated token) for this
# number of days
FACEBOOK_OG_SHARE_RETRY_DAYS = getattr(
    settings, 'FACEBOOK_OG_SHARE_RETRY_DAYS', 7)
FACEBOOK_OG_SHARE_DB_TABLE = getattr(
    settings, 'FACEBOOK_OG_SHARE_DB_TABLE', None)


# Force profile update every login
FACEBOOK_FORCE_PROFILE_UPDATE_ON_LOGIN = getattr(
    settings, 'FACEBOOK_FORCE_PROFILE_UPDATE_ON_LOGIN', False)

# Profile image location
FACEBOOK_PROFILE_IMAGE_PATH = getattr(
    settings, 'FACEBOOK_PROFILE_IMAGE_PATH', None)

# Ability to easily overwrite classes used for certain tasks
FACEBOOK_CLASS_MAPPING = getattr(
    settings, 'FACEBOOK_CLASS_MAPPING', None)

FACEBOOK_SKIP_VALIDATE = getattr(
    settings, 'FACEBOOK_SKIP_VALIDATE', False)



from django.dispatch import Signal


# Sent right after user is created
facebook_user_registered = Signal(providing_args=['user', 'facebook_data'])

# Sent after user is created, before profile is updated with data from Facebook
facebook_pre_update = Signal(
    providing_args=['user', 'profile', 'facebook_data'])
facebook_post_update = Signal(
    providing_args=['user', 'profile', 'facebook_data'])

# Sent after storing the friends from graph to db
facebook_post_store_friends = Signal(
    providing_args=['user', 'friends', 'current_friends', 'inserted_friends'])

# Sent after storing the likes from graph to db
facebook_post_store_likes = Signal(
    providing_args=['user', 'likes', 'current_likes', 'inserted_likes'])

# Some signals for compatibility with Django Registration
# A new user has registered.
user_registered = Signal(providing_args=['user', 'request'])

# A user has activated his or her account.
user_activated = Signal(providing_args=['user', 'request'])

# Run when the token extend finished
facebook_token_extend_finished = Signal(
    providing_args=['user', 'profile', 'token_changed', 'old_token'])



'''
Facebook error classes also see
http://fbdevwiki.com/wiki/Error_codes#User_Permission_Errors
'''
from __future__ import unicode_literals

import ssl

try:
    # python 2 imports
    from urllib2 import HTTPError, URLError
except ImportError:
    # python 3 imports
    from urllib.error import HTTPError, URLError


class OpenFacebookException(Exception):

    '''
    BaseClass for all open facebook errors
    '''

    @classmethod
    def codes_list(cls):
        '''
        Returns the codes as a list of instructions
        '''
        if hasattr(cls, 'codes'):
            codes_list = [cls.codes]
            if isinstance(cls.codes, list):
                codes_list = cls.codes
            return codes_list

    @classmethod
    def range(cls):
        '''
        Returns for how many codes this Exception, matches with the eventual
        goal of matching an error to the most specific error class
        '''
        range = 0
        codes_list = cls.codes_list()
        for c in codes_list:
            if isinstance(c, tuple):
                start, stop = c
                range += stop - start + 1
            else:
                range += 1

        # make sure none specific exceptions are last in the order
        if not range:
            range = 1000

        return range


class ParameterException(OpenFacebookException):

    '''
    100-189
    190 and up are oauth errors
    '''
    codes = (100, 189)


class UnknownException(OpenFacebookException):

    '''
    Raised when facebook themselves don't know what went wrong
    '''
    codes = 1


class OAuthException(OpenFacebookException):
    pass


class PermissionException(OAuthException):

    '''
    200-300
    '''
    codes = [3, (200, 299)]


class UserPermissionException(PermissionException):
    codes = (300, 399)


class FeedActionLimit(UserPermissionException):

    '''
    When you posted too many times from one user acount
    '''
    codes = 341


class OpenGraphException(OpenFacebookException):

    '''
    Raised when we get error 3502, representing a problem with facebook
    open graph data on the page
    '''
    codes = 3502


class DuplicateStatusMessage(OpenFacebookException):
    codes = 506


class MissingParameter(OpenFacebookException):
    pass


class AliasException(OpenFacebookException):

    '''
    When you send a request to a non existant url facebook gives this error
    instead of a 404....
    '''
    codes = 803


class UnsupportedDeleteRequest(OpenFacebookException):
    pass


class ParseException(OpenFacebookException):

    '''
    Anything preventing us from parsing the Facebook response
    '''
    pass


class FacebookUnreachable(OpenFacebookException):

    '''
    Timeouts, 500s, SSL errors etc
    '''
    pass


class FacebookSSLError(FacebookUnreachable, ssl.SSLError):
    pass


class FacebookHTTPError(FacebookUnreachable, HTTPError):
    pass


class FacebookURLError(FacebookUnreachable, URLError):
    pass


def map_unreachable_exception(e):
    '''
    We always raise the original and new subclass to
     - preserve backwards compatibility
    '''
    exception_class = FacebookUnreachable
    if isinstance(e, ssl.SSLError):
        exception_class = FacebookSSLError
    elif isinstance(e, HTTPError):
        exception_class = FacebookHTTPError
    elif isinstance(e, URLError):
        exception_class = FacebookURLError
    return exception_class


def convert_unreachable_exception(e, error_format='Facebook is unreachable %s'):
    '''
    Converts an SSLError, HTTPError or URLError into something subclassing
    FacebookUnreachable allowing code to easily try except this
    '''
    exception_class = map_unreachable_exception(e)
    error_message = error_format % str(e)
    exception = exception_class(error_message)
    return exception


def get_exception_classes():
    from open_facebook import exceptions as facebook_exceptions
    all_exceptions = dir(facebook_exceptions)
    classes = [getattr(facebook_exceptions, e, None) for e in all_exceptions]
    exception_classes = [e for e in classes if getattr(
                         e, 'codes', None) and issubclass(
                         e, OpenFacebookException)]
    return exception_classes



from django.conf import settings
from django.contrib import messages
from django.http import Http404
from django.shortcuts import render
from django.utils.translation import ugettext as _
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.http import require_POST
from django_facebook import exceptions as facebook_exceptions, \
    settings as facebook_settings
from django_facebook.connect import CONNECT_ACTIONS, connect_user
from django_facebook.decorators import facebook_required_lazy
from django_facebook.utils import next_redirect, get_registration_backend, \
    to_bool, error_next_redirect, get_instance_for, get_profile
from open_facebook import exceptions as open_facebook_exceptions
from open_facebook.utils import send_warning
import logging

try:
    unicode = unicode
except NameError:
    unicode = str

logger = logging.getLogger(__name__)


@csrf_exempt
@facebook_required_lazy
def connect(request, graph):
    '''
    Exception and validation functionality around the _connect view
    Separated this out from _connect to preserve readability
    Don't bother reading this code, skip to _connect for the bit you're interested in :)
    '''
    backend = get_registration_backend()

    try:
        response = _connect(request, graph)
    except open_facebook_exceptions.FacebookUnreachable as e:
        # often triggered when Facebook is slow
        warning_format = u'%s, often caused by Facebook slowdown, error %s'
        warn_message = warning_format % (type(e), str(e))
        send_warning(warn_message, e=e)
        additional_params = dict(fb_error_or_cancel=1)
        response = backend.post_error(request, additional_params)

    return response


def _connect(request, graph):
    '''
    Handles the view logic around connect user
    - (if authenticated) connect the user
    - login
    - register

    We are already covered by the facebook_required_lazy decorator
    So we know we either have a graph and permissions, or the user denied
    the oAuth dialog
    '''
    backend = get_registration_backend()
    connect_facebook = to_bool(
        request.POST.get(
            'connect_facebook',
            request.GET.get('connect_facebook'))
    )

    logger.info('trying to connect using Facebook')
    if graph:
        logger.info('found a graph object')
        converter = get_instance_for('user_conversion', graph)
        authenticated = converter.is_authenticated()
        # Defensive programming :)
        if not authenticated:
            raise ValueError('didnt expect this flow')

        logger.info('Facebook is authenticated')
        facebook_data = converter.facebook_profile_data()
        # either, login register or connect the user
        try:
            action, user = connect_user(
                request, connect_facebook=connect_facebook)
            logger.info('Django facebook performed action: %s', action)
        except facebook_exceptions.IncompleteProfileError as e:
            # show them a registration form to add additional data
            warning_format = u'Incomplete profile data encountered with error %s'
            warn_message = warning_format % unicode(e)
            send_warning(warn_message, e=e,
                         facebook_data=facebook_data)

            context = {'facebook_mode': True, 'form': e.form}
            return render(request, backend.get_registration_template(), context)
        except facebook_exceptions.AlreadyConnectedError as e:
            user_ids = [u.get_user_id() for u in e.users]
            ids_string = ','.join(map(str, user_ids))
            additional_params = dict(already_connected=ids_string)
            return backend.post_error(request, additional_params)

        response = backend.post_connect(request, user, action)

        if action is CONNECT_ACTIONS.LOGIN:
            pass
        elif action is CONNECT_ACTIONS.CONNECT:
            # connect means an existing account was attached to facebook
            messages.info(request, _("You have connected your account "
                                     "to %s's facebook profile") % facebook_data['name'])
        elif action is CONNECT_ACTIONS.REGISTER:
            # hook for tying in specific post registration functionality
            response.set_cookie('fresh_registration', user.id)
    else:
        # the user denied the request
        additional_params = dict(fb_error_or_cancel='1')
        response = backend.post_error(request, additional_params)

    return response


def disconnect(request):
    '''
    Removes Facebook from the users profile
    And redirects to the specified next page
    '''
    if request.method == 'POST':
        messages.info(
            request, _("You have disconnected your Facebook profile."))
        profile = get_profile(request.user)
        profile.disconnect_facebook()
        profile.save()
    response = next_redirect(request)
    return response


def example(request):
    context = {}
    return render(request, 'django_facebook/example.html', context)



from django.contrib.auth import backends
from django.db.models.query_utils import Q
from django.db.utils import DatabaseError
from django_facebook import settings as facebook_settings
from django_facebook.utils import get_profile_model, is_user_attribute, \
    get_user_model, get_profile
import operator

try:
    reduce = reduce
except NameError:
    from functools import reduce


class FacebookBackend(backends.ModelBackend):

    '''
    Django Facebook authentication backend

    This backend hides the difference between authenticating with
    - a django 1.5 custom user model
    - profile models, which were used prior to 1.5

    **Example usage**

    >>> FacebookBackend().authenticate(facebook_id=myid)

    '''

    def authenticate(self, facebook_id=None, facebook_email=None):
        '''
        Route to either the user or profile table depending on which type of user
        customization we are using
        (profile was used in Django < 1.5, user is the new way in 1.5 and up)
        '''
        user_attribute = is_user_attribute('facebook_id')
        if user_attribute:
            user = self.user_authenticate(facebook_id, facebook_email)
        else:
            user = self.profile_authenticate(facebook_id, facebook_email)
        return user

    def user_authenticate(self, facebook_id=None, facebook_email=None):
        '''
        Authenticate the facebook user by id OR facebook_email
        We filter using an OR to allow existing members to connect with
        their facebook ID using email.

        This decorator works with django's custom user model

        :param facebook_id:
            Optional string representing the facebook id.
        :param facebook_email:
            Optional string with the facebook email.

        :return: The signed in :class:`User`.
        '''
        user_model = get_user_model()
        if facebook_id or facebook_email:
            # match by facebook id or email
            auth_conditions = []
            if facebook_id:
                auth_conditions.append(Q(facebook_id=facebook_id))
            if facebook_email:
                auth_conditions.append(Q(email__iexact=facebook_email))
            # or the operations
            auth_condition = reduce(operator.or_, auth_conditions)

            # get the users in one query
            users = list(user_model.objects.filter(auth_condition))

            # id matches vs email matches
            id_matches = [u for u in users if u.facebook_id == facebook_id]
            email_matches = [u for u in users if u.facebook_id != facebook_id]

            # error checking
            if len(id_matches) > 1:
                raise ValueError(
                    'found multiple facebook id matches. this shouldnt be allowed, check your unique constraints. users found %s' % users)

            # give the right user
            user = None
            if id_matches:
                user = id_matches[0]
            elif email_matches:
                user = email_matches[0]

            if user and facebook_settings.FACEBOOK_FORCE_PROFILE_UPDATE_ON_LOGIN:
                user.fb_update_required = True
            return user

    def profile_authenticate(self, facebook_id=None, facebook_email=None):
        '''
        Authenticate the facebook user by id OR facebook_email
        We filter using an OR to allow existing members to connect with
        their facebook ID using email.

        :param facebook_id:
            Optional string representing the facebook id

        :param facebook_email:
            Optional string with the facebook email

        :return: The signed in :class:`User`.
        '''
        if facebook_id or facebook_email:
            profile_class = get_profile_model()
            profile_query = profile_class.objects.all().order_by('user')
            profile_query = profile_query.select_related('user')
            profile = None

            # filter on email or facebook id, two queries for better
            # queryplan with large data sets
            if facebook_id:
                profiles = profile_query.filter(facebook_id=facebook_id)[:1]
                profile = profiles[0] if profiles else None
            if profile is None and facebook_email:
                try:
                    profiles = profile_query.filter(
                        user__email__iexact=facebook_email)[:1]
                    profile = profiles[0] if profiles else None
                except DatabaseError:
                    try:
                        user = get_user_model(
                        ).objects.get(email=facebook_email)
                    except get_user_model().DoesNotExist:
                        user = None
                    profile = get_profile(user) if user else None

            if profile:
                # populate the profile cache while we're getting it anyway
                user = profile.user
                user._profile = profile
                if facebook_settings.FACEBOOK_FORCE_PROFILE_UPDATE_ON_LOGIN:
                    user.fb_update_required = True
                return user



class FacebookException(Exception):

    '''
    Base class for Facebook related exceptions
    '''
    pass


class IncompleteProfileError(FacebookException):

    '''
    Raised when we get insufficient data to create a profile for a user.
    One example is a Facebook token, without permissions to see the email.
    '''
    pass


class AlreadyConnectedError(FacebookException):

    '''
    Raised when another user account is already connected to your Facebook id
    '''

    def __init__(self, users):
        self.users = users


class AlreadyRegistered(FacebookException):

    '''
    Raised if you try to register when there's already an account with
    the given email or facebook id
    '''
    pass


class MissingPermissionsError(FacebookException):

    '''
    Raised if we lack permissions
    '''
    pass
